# coding=utf-8

from unittest import TestCase, skip

from abilian.sbe.apps.documents.models import Folder, Document, icon_for


def check_editable(object):
  if hasattr(object, '__editable__'):
    for k in object.__editable__:
      assert hasattr(object, k)


def test_title_prevails():
  f = Folder(name=u'name', title=u'title')
  assert f.title == u'title'
  assert f.name == u'title'

  f = Folder(name=u'name', title=None)
  assert f.title == u'name'
  assert f.name == u'name'

  f = Folder(name=u'name')
  assert f.title == u'name'
  assert f.name == u'name'


def test_folder_editables():
  root = Folder(title="/")
  check_editable(root)

def test_folder_can_create_documents():
  root = Folder(title="/")

  document = root.create_document("doc")

  assert len(root.children) == 1
  assert document.title == "doc"
  assert document.path == "/doc"
  assert document in root.documents
  assert document in root.children
  assert document.parent == root

def test_folder_can_create_subfolders():
  root = Folder(title="/")

  subfolder = root.create_subfolder("folder")
  assert len(root.children) == 1
  assert subfolder.title == "folder"
  assert subfolder.path == "/folder"
  assert subfolder in root.subfolders
  assert subfolder in root.children
  assert subfolder.parent == root

def test_nested_subobjects():
  root = Folder(title="/")
  subfolder = root.create_subfolder("folder1")
  subsubfolder = subfolder.create_subfolder("folder2")
  document = subfolder.create_document("doc")

  assert len(root.children) == 1
  assert len(subfolder.children) == 2
  assert subsubfolder.title == "folder2"
  assert subsubfolder.path == "/folder1/folder2"
  assert document.title == "doc"
  assert document.path == "/folder1/doc"

  assert root.get_object_by_path("/folder1") == subfolder
  assert root.get_object_by_path("/folder1/folder2") == subsubfolder
  assert root.get_object_by_path("/folder1/doc") == document

def test_folder_is_clonable():
  root = Folder(title="/")
  clone = root.clone()

  assert clone.title == root.title
  assert clone.path == root.path


# FIXME: skipped for now as these tests break other tests down the line (in wiki).
@skip
class TestDocument(TestCase):

  def test_document_editables(self):
    doc = Document()
    check_editable(doc)

  def test_content_length(self):
    doc = Document()
    doc.set_content("tototiti", "application/binary")
    assert doc.content_length == len("tototiti")

  def test_document_is_clonable(self):
    root = Folder(title="/")
    doc = root.create_document(title="toto")
    doc.content = "tototiti"

    clone = doc.clone()
    assert clone.title == doc.title

  def test_document_has_an_icon(self):
    root = Folder(title="/")
    doc = root.create_document(title="toto")
    doc.content_type = "image/jpeg"
    filename = doc.icon.split("/")[-1]
    assert filename in ("jpg.png", "jpeg.png"), doc.icon

  def test_icon_from_mime_type(self):
    icon = icon_for("text/html")
    filename = icon.split("/")[-1]
    assert filename in ("html.png", "htm.png"), icon



"""
Decorators and helpers to check access to communities.
"""

from functools import wraps

from flask import g, current_app
from flask_login import current_user
from werkzeug.exceptions import Forbidden


def require_admin(func):
  @wraps(func)
  def decorated_view(*args, **kwargs):
    security = current_app.services['security']
    is_admin = security.has_role(current_user, 'admin')
    if not is_admin:
      raise Forbidden()
    return func(*args, **kwargs)

  return decorated_view


def require_manage(func):
  @wraps(func)
  def decorated_view(*args, **kwargs):
    community = getattr(g, 'community')
    if community and community.has_permission(current_user, 'manage'):
      return func(*args, **kwargs)
    security = current_app.services['security']
    is_admin = security.has_role(current_user, 'admin')
    if not is_admin:
      raise Forbidden()
    return func(*args, **kwargs)

  return decorated_view


def require_access(func):
  @wraps(func)
  def decorated_view(*args, **kwargs):
    check_access()
    return func(*args, **kwargs)

  return decorated_view


def check_access(community=None, user=None):
  if not has_access(community, user):
    raise Forbidden()


def has_access(community=None, user=None):
  if not user:
    user = current_user
  if user.is_anonymous():
    return False

  security = current_app.services['security']
  is_admin = security.has_role(user, 'admin')
  if is_admin:
    return True

  if not community:
    community = getattr(g, 'community', None)

  if community is not None:
    return community.get_role(user) is not None

  return False



# coding=utf-8
"""
Set up the markdown converter. Add extensions here (for now).
"""
from __future__ import absolute_import

from flask import url_for
import markdown
from markdown.extensions.wikilinks import WikiLinkExtension, WikiLinks
from markdown.util import etree
__all__ = ['convert']


class UrlBuilder(object):
  def __init__(self, page):
    self.page = page

  def build(self, label, base, end):
    label = label.strip()
    return url_for(".page",
                   community_id=self.page.community.slug,
                   title=label)


def convert(page, text):
  build_url = UrlBuilder(page).build
  extension = SBEWikiLinkExtension([('build_url', build_url)])
  ctx = {}
  ctx['extensions'] = [extension, 'toc']
  ctx['output_format'] = 'html5'
  md = markdown.Markdown(**ctx)
  return md.convert(text)


class SBEWikiLinkExtension(WikiLinkExtension):
  def extendMarkdown(self, md, md_globals):
    self.md = md

    # append to end of inline patterns
    WIKILINK_RE = r'\[\[(.*?)\]\]'
    wikilinkPattern = SBEWikiLinks(WIKILINK_RE, self.getConfigs())
    wikilinkPattern.md = md
    md.inlinePatterns.add('wikilink', wikilinkPattern, "<not_strong")


class SBEWikiLinks(WikiLinks):
  def handleMatch(self, m):
    from .forms import page_exists
    if m.group(2).strip():
        base_url, end_url, html_class = self._getMeta()
        label = m.group(2).strip()
        url = self.config['build_url'](label, base_url, end_url)
        a = etree.Element('a')
        a.text = label
        a.set('href', url)
        if html_class:
          if page_exists(label):
            a.set('class', html_class)
          else:
            a.set('class', html_class + ' new')
    else:
        a = ''
    return a



from __future__ import print_function

from abilian.web import url_for
from abilian.sbe.testing import BaseTestCase


class TestViews(BaseTestCase):
  init_data = True
  no_login = True

  def test_all_registered_urls(self):

    SKIP = frozenset([
        'admin.audit_search_users',
        'search.search_main', 'search.live',
        'notifications.debug_social',
        'social.groups_json', 'social.groups_new',
        'social.users_json', 'social.users_dt_json',
        'communities.community_default_image',
        'images.user_default',
    ])

    rules = sorted(self.app.url_map.iter_rules(), key=lambda x: x.endpoint)
    for rule in rules:

      if "GET" not in rule.methods:
        continue

      if rule.arguments:
        continue

      # FIXME. Skip several exceptions.
      if rule.endpoint in SKIP:
        continue

      # These endpoints expect a parameter ('q') that we don't provide
      if rule.endpoint.endswith("list_json2"):
        continue

      url = url_for(rule.endpoint)

      if '/ajax/' in url:
        continue

      print(rule.endpoint, url)

      try:
        response = self.client.get(url)
        assert response.status_code in (200, 302), \
          "Bad link: {} (status={})".format(url, response.status_code)
      except:
        print("Problem with url: {}".format(url))
        raise



# coding=utf-8
"""
Notifications
"""
from __future__ import absolute_import

from abilian.sbe.extension import sbe

# Constants
TOKEN_SERIALIZER_NAME = "unsubscribe_sbe"


def register_plugin(app):
  cfg = app.config.setdefault('ABILIAN_SBE', {})
  cfg.setdefault('DAILY_SOCIAL_DIGEST_SUBJECT',
                 u'Des nouvelles de vos communautés')
  sbe.init_app(app)

  # TODO: Slightly confusing. Reorg?
  from .views import notifications, social  # noqa
  from .tasks import social  # noqa

  app.register_blueprint(notifications)



from flask import render_template

# TEMP
ROOT = "http://localhost:5000/cmis/atompub"
XML_HEADER = "<?xml version='1.0' encoding='UTF-8'?>\n"


class Feed(object):
  def __init__(self, object, collection):
    self.object = object
    self.collection = collection

  def to_xml(self, **options):
    ctx = {'ROOT': ROOT, 'object': self.object,
           'collection': self.collection, 'to_xml': to_xml}
    return render_template("cmis/feed.xml", **ctx)


class Entry(object):
  def __init__(self, obj):
    self.obj = obj

  def to_xml(self, **options):
    ctx = {'ROOT': ROOT, 'folder': self.obj, 'document': self.obj,
           'options': options, 'to_xml': to_xml}

    if self.obj.sbe_type == 'cmis:folder':
      result = render_template("cmis/folder.xml", **ctx)
    elif self.obj.sbe_type == 'cmis:document':
      result = render_template("cmis/document.xml", **ctx)
    else:
      raise Exception("Unknown base object type: %s" % self.obj.sbe_type)

    if not 'no_xml_header' in options:
      result = XML_HEADER + result

    return result


def to_xml(obj, **options):
  entry = Entry(obj)
  return entry.to_xml(**options)



"""
Parses XML messages and converts them to objects.

"""
import base64

from datetime import datetime
from lxml import objectify

ATOM_NS = "http://www.w3.org/2005/Atom"
APP_NS = "http://www.w3.org/2007/app"
CMISRA_NS = "http://docs.oasis-open.org/ns/cmis/restatom/200908/"
CMIS_NS = "http://docs.oasis-open.org/ns/cmis/core/200908/"


class Entry(object):

  def __init__(self, xml=None):
    self.properties = {}
    self.links = []
    self.content = None
    self.content_type = None

    if xml:
      self.parse(xml)

  def parse(self, xml):
    root = objectify.fromstring(xml)
    object = root['{%s}object' % CMISRA_NS]
    properties = object['{%s}properties' % CMIS_NS]
    for element in properties.iterchildren():
      property = Property(element)
      self.properties[property.property_definition_id] = property

    content_element = getattr(root, '{%s}content' % CMISRA_NS, None)
    if content_element is not None:
      self.content_type = content_element.mediatype.text
      self.content = base64.b64decode(content_element.base64.text)

  @property
  def name(self):
    return self.properties['cmis:name'].value

  @property
  def type(self):
    return self.properties['cmis:objectTypeId'].value


class Property(object):
  """
  A property MAY hold zero, one, or more typed data value(s).
  Each property MAY be single-valued or multi-valued. A single-valued property
  contains a single data value, whereas a multi-valued property contains an
  ordered list of data values of the same type. The ordering of values in a
  multi-valued property SHOULD be preserved by the repository.

  A property, either single-valued or multi-valued, MAY be in a "not set" state.
  CMIS does not support "null" property value. If a multi-valued property
  is not in a "not set" state, its property value MUST be a non-empty list of
  individual values. Each individual value in the list MUST NOT be in a "not
  set" state and MUST conform to the property's property-type.

  A multi-valued property is either set or not set in its entirety. An
  individual value of a multi-valued property MUST NOT be in an individual
  "value not set" state and hold a position in the list of values. An empty
  list of values MUST NOT be allowed.

  Every property is typed. The property-type defines the data type of the data
  value(s) held by the property. CMIS specifies the following property-types.
  """

  def __init__(self, element=None):
    if element is not None:
      self.parse(element)

  def parse(self, element):
    tag = element.tag
    self.type = tag[tag.index("}") + 1 + len("property"):].lower()
    self.property_definition_id = element.attrib['propertyDefinitionId']
    self.local_name = element.attrib.get('localName')
    self.display_name = element.attrib.get('displayName')
    self.query_name = element.attrib.get('queryName')

    value_elem = getattr(element, "{%s}value" % CMIS_NS)
    if value_elem:
      value = value_elem.text
    else:
      self.value = value = None

    if value:
      if self.type in ('id', 'string'):
        self.value = value
      elif self.type == 'datetime':
        # FIXME
        self.value = datetime(value)
      else:
        raise Exception("Unknown value type: %s" % self.type)



# coding=utf-8
"""
Forum views
"""
from __future__ import absolute_import, print_function

from datetime import datetime, date
from itertools import groupby
from urllib import quote

import sqlalchemy as sa
from sqlalchemy.orm import joinedload
from werkzeug.exceptions import NotFound, BadRequest
from flask import (
  current_app, g, make_response, render_template, request, flash,
)
from flask_babel import format_date
from flask_login import current_user

from abilian.i18n import _, _l
from abilian.core.util import utc_dt
from abilian.sbe.apps.communities.blueprint import Blueprint
from abilian.sbe.apps.communities.views import default_view_kw
from abilian.web import nav, url_for, views
from abilian.web.action import ButtonAction
from abilian.web.views import default_view

from .forms import PostForm, ThreadForm, PostEditForm
from .models import Post, PostAttachment, Thread
from .tasks import send_post_by_email

# TODO: move to config
MAX_THREADS = 30

forum = Blueprint("forum", __name__,
                  url_prefix="/forum",
                  template_folder="templates")
route = forum.route

def post_kw_view_func(kw, obj, obj_type, obj_id, **kwargs):
  """
  kwargs for Post default view
  """
  kw = default_view_kw(kw, obj.thread, obj_type, obj_id, **kwargs)
  kw['thread_id'] = obj.thread_id
  kw['_anchor'] = u'post_{:d}'.format(obj.id)
  return kw


@forum.url_value_preprocessor
def init_forum_values(endpoint, values):
  g.current_tab = 'forum'

  g.breadcrumb.append(
    nav.BreadcrumbItem(label=_l(u'Conversations'),
                       url=nav.Endpoint('forum.index',
                                        community_id=g.community.slug)))


@route('/')
def index():
  query = Thread.query \
    .filter(Thread.community_id == g.community.id) \
    .order_by(Thread.created_at.desc())
  has_more = query.count() > MAX_THREADS
  threads = query.limit(MAX_THREADS).all()
  return render_template("forum/index.html",
                         threads=threads, has_more=has_more)


def group_monthly(entities_list):
  # We're using Python's groupby instead of SA's group_by here
  # because it's easier to support both SQLite and Postgres this way.
  def grouper(entity):
    return entity.created_at.year, entity.created_at.month

  def format_month(year, month):
    month = format_date(date(year, month, 1), "MMMM").capitalize()
    return u"%s %s" % (month, year)

  grouped_entities = groupby(entities_list, grouper)
  grouped_entities = [(format_month(year, month), list(entities))
                      for (year, month), entities in grouped_entities]
  return grouped_entities


@route('/archives/')
def archives():
  all_threads = Thread.query \
    .filter(Thread.community_id == g.community.id) \
    .order_by(Thread.created_at.desc()).all()

  grouped_threads = group_monthly(all_threads)
  return render_template('forum/archives.html',
                         grouped_threads=grouped_threads)


@route('/attachments/')
def attachments():
  # XXX: there is probably a way to optimize this and the big loop below...
  all_threads = Thread.query \
    .filter(Thread.community_id == g.community.id) \
    .options(joinedload('posts')) \
    .options(joinedload('posts.attachments')) \
    .order_by(Thread.created_at.desc()).all()

  posts_with_attachments = []
  for thread in all_threads:
    for post in thread.posts:
      if getattr(post, 'attachments', None):
        posts_with_attachments.append(post)
  posts_with_attachments.sort(key=lambda post: post.created_at)
  posts_with_attachments.reverse()

  grouped_posts = group_monthly(posts_with_attachments)
  return render_template('forum/attachments.html',
                         grouped_posts=grouped_posts)


class BaseThreadView(object):
  Model = Thread
  Form = ThreadForm
  pk = 'thread_id'
  base_template = 'community/_base.html'

  def can_send_by_mail(self):
    return (g.community.type == 'participative'
            or g.community.has_permission(current_user, 'manage'))

  def prepare_args(self, args, kwargs):
    args, kwargs = super(BaseThreadView, self).prepare_args(args, kwargs)
    self.send_by_email = False

    if not self.can_send_by_mail() and 'send_by_email' in self.form:
      # remove from html form and avoid validation errors
      del self.form['send_by_email']

    return args, kwargs

  def index_url(self):
    return url_for(".index", community_id=g.community.slug)

  def view_url(self):
    return url_for(self.obj)


class ThreadView(BaseThreadView, views.ObjectView):
  methods = ['GET', 'HEAD']
  Form = PostForm
  template = 'forum/thread.html'

  @property
  def template_kwargs(self):
    kw = super(ThreadView, self).template_kwargs
    kw['thread'] = self.obj
    kw['is_closed'] = self.obj.closed
    return kw


thread_view = ThreadView.as_view('thread')
default_view(forum, Post, None, kw_func=post_kw_view_func)(thread_view)
default_view(forum, Thread, 'thread_id', kw_func=default_view_kw)(thread_view)

route('/<int:thread_id>/')(thread_view)
route('/<int:thread_id>/attachments')(
  ThreadView.as_view('thread_attachments',
                     template='forum/thread_attachments.html')
)


class ThreadCreate(BaseThreadView, views.ObjectCreate):
  POST_BUTTON = ButtonAction('form', 'create', btn_class='primary',
                             title=_l(u'Post this message'))

  def init_object(self, args, kwargs):
    args, kwargs = super(ThreadCreate, self).init_object(args, kwargs)
    self.thread = self.obj
    return args, kwargs

  def before_populate_obj(self):
    del self.form['attachments']
    self.message_body = self.form.message.data
    del self.form['message']

    if 'send_by_email' in self.form:
      self.send_by_email = (self.can_send_by_mail()
                            and self.form.send_by_email.data)
      del self.form['send_by_email']

  def after_populate_obj(self):
    if self.thread.community is None:
      self.thread.community = g.community._model

    self.post = self.thread.create_post(body_html=self.message_body)
    obj_meta = self.post.meta.setdefault('abilian.sbe.forum', {})
    obj_meta['origin'] = u'web'
    obj_meta['send_by_email'] = self.send_by_email

    session = sa.orm.object_session(self.thread)
    uploads = current_app.extensions['uploads']

    for handle in request.form.getlist('attachments'):
      fileobj = uploads.get_file(current_user, handle)
      if fileobj is None:
        continue

      meta = uploads.get_metadata(current_user, handle)
      name = meta.get('filename', handle)
      mimetype = meta.get('mimetype', None)

      if not isinstance(name, unicode):
        name = unicode(name, encoding='utf-8', errors='ignore')

      if not name:
        continue

      attachment = PostAttachment(name=name)
      attachment.post = self.post

      with fileobj.open('rb') as f:
        attachment.set_content(f.read(), mimetype)
      session.add(attachment)

  def commit_success(self):
    if self.send_by_email:
      task = send_post_by_email.delay(self.post.id)
      meta = self.post.meta.setdefault('abilian.sbe.forum', {})
      meta['send_post_by_email_task'] = task.id
      self.post.meta.changed()
      session = sa.orm.object_session(self.post)
      session.commit()

  @property
  def activity_target(self):
    return self.thread.community

  def get_form_buttons(self, *args, **kwargs):
    return [self.POST_BUTTON, views.object.CANCEL_BUTTON]


route('/new_thread/')(ThreadCreate.as_view('new_thread',
                                           view_endpoint='.thread'))


class ThreadPostCreate(ThreadCreate):
  """
  Add a new post to a thread
  """
  methods = ['POST']
  Form = PostForm
  Model = Post

  def init_object(self, args, kwargs):
    # we DO want to skip ThreadCreate.init_object. hence super is not based on
    # ThreadPostCreate
    args, kwargs = super(ThreadCreate, self).init_object(args, kwargs)
    thread_id = kwargs.pop(self.pk, None)
    self.thread = Thread.query.get(thread_id)
    return args, kwargs

  def after_populate_obj(self):
    super(ThreadPostCreate, self).after_populate_obj()
    session = sa.orm.object_session(self.obj)
    session.expunge(self.obj)
    self.obj = self.post


route('/<int:thread_id>/')(ThreadPostCreate.as_view('thread_post',
                                                    view_endpoint='.thread'))


class ThreadDelete(BaseThreadView, views.ObjectDelete):
  methods = ['POST']
  _message_success = _(u'Thread "{title}" deleted.')

  def message_success(self):
    return unicode(self._message_success).format(title=self.obj.title)


route('/<int:thread_id>/delete')(ThreadDelete.as_view('thread_delete'))


class ThreadCloseView(BaseThreadView, views.object.BaseObjectView):
  """
  Close / Re-open a thread
  """
  methods = ['POST']
  _VALID_ACTIONS = {u'close', u'reopen'}
  CLOSED_MSG = _l(u'The thread is now closed for edition and new '
                  u'contributions.')
  REOPENED_MSG = _l(u'The thread is now re-opened for edition and new '
                    u'contributions.')

  def prepare_args(self, args, kwargs):
    args, kwargs = super(ThreadCloseView, self).prepare_args(args, kwargs)
    action = kwargs['action'] = request.form.get('action')

    if action not in self._VALID_ACTIONS:
      raise BadRequest(u'Unknown action: {!r}'.format(action))

    return args, kwargs

  def post(self, action=None):
    is_closed = (action == u'close')
    self.obj.closed = is_closed
    sa.orm.object_session(self.obj).commit()

    msg = self.CLOSED_MSG if is_closed else self.REOPENED_MSG
    flash(unicode(msg))
    return self.redirect(url_for(self.obj))

route('/<int:thread_id>/close')(ThreadCloseView.as_view('thread_close'))


class ThreadPostEdit(BaseThreadView, views.ObjectEdit):
  Form = PostEditForm
  Model = Post
  pk = 'object_id'

  def can_send_by_mail(self):
    # post edit: don't notify every time
    return False

  def init_object(self, args, kwargs):
    # we DO want to skip ThreadCreate.init_object. hence super is not based on
    # ThreadPostCreate
    args, kwargs = super(ThreadPostEdit, self).init_object(args, kwargs)
    thread_id = kwargs.pop('thread_id', None)
    self.thread = self.obj.thread
    assert thread_id == self.thread.id
    return args, kwargs

  def get_form_kwargs(self):
    kwargs = super(ThreadPostEdit, self).get_form_kwargs()
    kwargs['message'] = self.obj.body_html
    return kwargs

  def before_populate_obj(self):
    self.message_body = self.form.message.data
    del self.form['message']
    self.reason = self.form.reason.data

    self.send_by_email = False
    if 'send_by_email' in self.form:
      del self.form['send_by_email']

    self.attachments_to_remove = self.form['attachments'].delete_files_index
    del self.form['attachments']

  def after_populate_obj(self):
    session = sa.orm.object_session(self.obj)
    uploads = current_app.extensions['uploads']
    self.obj.body_html = self.message_body
    obj_meta = self.obj.meta.setdefault('abilian.sbe.forum', {})
    history = obj_meta.setdefault('history', [])
    history.append(dict(user_id=current_user.id,
                        user=unicode(current_user),
                        date=utc_dt(datetime.utcnow()).isoformat(),
                        reason=self.form.reason.data,))
    self.obj.meta['abilian.sbe.forum'] = obj_meta # trigger change for SA

    attachments_to_remove = []
    for idx in self.attachments_to_remove:
      try:
        idx = int(idx)
      except ValueError:
        continue

      if idx > len(self.obj.attachments):
        continue

      attachments_to_remove.append(self.obj.attachments[idx])

    for att in attachments_to_remove:
      session.delete(att)

    for handle in request.form.getlist('attachments'):
      fileobj = uploads.get_file(current_user, handle)
      if fileobj is None:
        continue

      meta = uploads.get_metadata(current_user, handle)
      name = meta.get('filename', handle)
      mimetype = meta.get('mimetype', None)

      if not isinstance(name, unicode):
        name = unicode(name, encoding='utf-8', errors='ignore')

      if not name:
        continue

      attachment = PostAttachment(name=name, post=self.obj)

      with fileobj.open('rb') as f:
        attachment.set_content(f.read(), mimetype)
      session.add(attachment)


route('/<int:thread_id>/<int:object_id>/edit')(
  ThreadPostEdit.as_view('post_edit')
)

def attachment_kw_view_func(kw, obj, obj_type, obj_id, **kwargs):
  post = obj.post
  kw = default_view_kw(kw, post.thread, obj_type, obj_id, **kwargs)
  kw['thread_id'] = post.thread_id
  kw['post_id'] = post.id
  return kw


@route('/<int:thread_id>/posts/<int:post_id>/attachment/<int:attachment_id>')
@default_view(forum, PostAttachment, 'attachment_id',
              kw_func=attachment_kw_view_func)
def attachment_download(thread_id, post_id, attachment_id):
  thread = Thread.query.get(thread_id)
  post = Post.query.get(post_id)
  attachment = PostAttachment.query.get(attachment_id)

  if (not (thread and post and attachment)
      or post.thread is not thread
      or attachment.post is not post):
    raise NotFound()

  response = make_response(attachment.content)
  response.headers['content-length'] = attachment.content_length
  response.headers['content-type'] = attachment.content_type
  content_disposition = (
    'attachment;filename="{}"'.format(quote(attachment.name.encode('utf8')))
  )
  response.headers['content-disposition'] = content_disposition
  return response



import imghdr
from string import strip

import PIL
from flask import request
from flask_babel import lazy_gettext as _l, gettext as _
from wtforms.fields import BooleanField, TextAreaField, StringField
from wtforms.validators import ValidationError, required, optional
from abilian.web.forms import Form
from abilian.web.forms.fields import Select2Field, FileField
from abilian.web.forms.widgets import TextArea, ImageInput, BooleanWidget
from abilian.web.forms.validators import length

from .models import Community


class CommunityForm(Form):
  name = StringField(label=_l(u"Name"), validators=[required()])
  description = TextAreaField(
      label=_l(u"Description"),
      validators=[required(), length(max=500)],
      widget=TextArea(resizeable="vertical"),)

  image = FileField(label=_l('Image'), widget=ImageInput(width=65, height=65),
                    validators=[optional()])

  type = Select2Field(label=_(u"Type"), validators=[required()],
                      filters=(strip,),
                      choices=[(_l(u'informative'), 'informative'),
                               (_l(u'participative'), 'participative')])

  has_documents = BooleanField(label=_l(u"Has documents"),
                               widget=BooleanWidget(on_off_mode=True))
  has_wiki = BooleanField(label=_l(u"Has a wiki"),
                          widget=BooleanWidget(on_off_mode=True))
  has_forum = BooleanField(label=_l(u"Has a forum"),
                           widget=BooleanWidget(on_off_mode=True))

  def validate_name(self, field):
    name = field.data = field.data.strip()

    if name and field.object_data:
      # form is bound to an existing object, name is not empty
      if name != field.object_data:
        # name changed: check for duplicates
        if len(list(Community.query.filter(Community.name==name).values('id'))) > 0:
          raise ValidationError(_(u"A community with this name already exists"))

  def validate_description(self, field):
    field.data = field.data.strip()

  # FIXME: code duplicated from the user edit form (UserProfileForm).
  # Needs to be refactored.
  def validate_image(self, field):
    data = request.form.get('image')
    if not data:
      return

    data = field.data
    filename = data.filename
    valid = any(map(filename.lower().endswith, ('.png', '.jpg', '.jpeg')))

    if not valid:
      raise ValidationError(_(u'Only PNG or JPG image files are accepted'))

    img_type = imghdr.what('ignored', data.read())

    if not img_type in ('png', 'jpeg'):
      raise ValidationError(_(u'Only PNG or JPG image files are accepted'))

    data.seek(0)
    try:
      # check this is actually an image file
      im = PIL.Image.open(data)
      im.load()
    except:
      raise ValidationError(_(u'Could not decode image file'))

    data.seek(0)
    field.data = data



from abilian.core.models.subjects import User
from abilian.sbe.apps.communities.models import Community, READER
from abilian.sbe.testing import BaseTestCase


class CommunityBaseTestCase(BaseTestCase):

  no_login = True

  def setUp(self):
    super(CommunityBaseTestCase, self).setUp()
    self.community = Community(name=u"My Community")
    self.session.add(self.community)
    self.session.flush()


class CommunityIndexingTestCase(CommunityBaseTestCase):
  """
  Testcase for testing indexing and searching of community content.
  """

  no_login = False
  SERVICES = ('security', 'indexing',)

  def setUp(self):
    super(CommunityIndexingTestCase, self).setUp()
    self.svc = self.app.services['indexing']
    self.user = User(email=u'user_1@example.com', password='azerty',
                     can_login=True)
    self.session.add(self.user)
    self.community.set_membership(self.user, READER)
    self.c2 = Community(name=u'Other community')
    self.session.add(self.c2)
    self.user_c2 = User(email=u'user_2@example.com', password='azerty',
                        can_login=True)
    self.session.add(self.user_c2)
    self.c2.set_membership(self.user_c2, READER)

    self.user_no_community = User(email=u'no_community@example.com',
                                  password='azerty',
                                  can_login=True)
    self.session.add(self.user_no_community)
    self.session.commit()



# coding=utf-8
"""
"""
from __future__ import absolute_import

import sqlalchemy as sa
from flask_script import Manager

from . import tasks
from .models import Document

manager = Manager(description='SBE documents actions',
                  help='SBE documents actions')

@manager.command
def antivirus():
  """
  Schedule documents to antivirus scan.
  """
  documents = Document.query\
      .filter(Document.content_blob != None)\
      .options(sa.orm.noload('creator'),
               sa.orm.noload('owner'),
               sa.orm.joinedload('content_blob')
      )

  total = 0
  count = 0
  for d in documents.yield_per(1000):
    total += 1
    meta = d.content_blob.meta
    if 'antivirus' not in meta and 'antivirus_task' not in meta:
      tasks.antivirus_scan.delay(d.id)
      count += 1

  print('{count}/{total} documents scheduled'.format(count=count, total=total))



# encoding: utf-8
"""
Test the application using py.test and splinter.

See: https://splinter.readthedocs.org/

Also, we're using pytest-splinter
(https://github.com/pytest-dev/pytest-splinter) to inject the browser as
a pytest fixture.
"""
import socket
import tempfile
import shutil
import multiprocessing
import pytest
import pytest_splinter.plugin  # noqa
from werkzeug.serving import select_ip_version

from abilian.sbe.app import Application


@pytest.fixture(scope='session')
def app_port(request):
  port = 30000
  hostname = '127.0.0.1'

  address_family = select_ip_version(hostname, port)
  test_socket = socket.socket(address_family, socket.SOCK_STREAM)
  test_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
  while port < 60000:
    try:
      test_socket.bind((hostname, port))
      test_socket.close()

    except socket.error as exc:
      if exc.errno != 98:
        # errno(98, Address already in use)
        raise
      port +=1

    break

  return port


@pytest.fixture(scope='session')
def app_root(app_port):
  return "http://localhost:{}".format(app_port)


# We're using phantomjs as our default browser.
@pytest.fixture(scope='session')  # pragma: no cover
def splinter_webdriver(request):
  return 'phantomjs'


@pytest.fixture(scope='session')
def splinter_driver_kwargs():
    """Webdriver kwargs."""
    # set resourceTimeout in the hope it will help jenkins kill phantomjs after
    # tests
    return dict(
      desired_capabilities={
        'phantomjs.page.settings.resourceTimeout': '30000', # 30s
      },
    )


@pytest.fixture(scope='module')
def instance_path(request):
  """
  creates a temporary directory for instance data
  """
  tmp_dir = tempfile.mkdtemp(prefix='tmp-pytest-', suffix='-abilian-sbe')

  def clear():
    shutil.rmtree(tmp_dir)

  request.addfinalizer(clear)
  return tmp_dir


@pytest.fixture(scope='module')
def app(request, instance_path, app_port):
  app = Application(instance_path=instance_path)

  # FIXME: need a working environment (DB, Redis...)
  # with app.app_context():
  #   app.create_db()

  def worker(app, port):
    app.run(port=port)

  process = multiprocessing.Process(target=worker, args=(app, app_port))

  try:
    process.start()
  except Exception as e:
    pytest.fail(e.message)

  def finalizer():
    if process:
      process.terminate()

  request.addfinalizer(finalizer)

  return app


def test_home(browser, app, app_root):
  browser.visit(app_root)


def test_login(browser, app, app_root):
  browser.visit(app_root + '/user/login')
  browser.fill('email', "admin@example.com")
  browser.fill('password', "admin")
  button = browser.find_by_xpath("//form[@name='login']//button")
  assert button

  # FIXME: at this point this raises an error because we have to set up the db.
  # button.click()
  # assert "Welcome to Abilian" in browser.find_element_by_xpath("/html/body").text


def test_forgotten_pw(browser, app, app_root):
  browser.visit(app_root + '/user/forgotten_pw')



from flask_babel import lazy_gettext as _l
from abilian.core.util import BasePresenter


class CommunityPresenter(BasePresenter):
  @property
  def breadcrumbs(self):
    return [dict(label=_l("Communities"), path="/communities/"),
            dict(label=self._model.name)]



# coding=utf-8
from flask import g, url_for as url_for_orig
from flask_babel import lazy_gettext as _l
from flask_login import current_user
from abilian.web.action import actions, Action, Endpoint
from abilian.web.nav import NavItem

__all__ = ['register_actions']


def url_for(endpoint, **kw):
  return url_for_orig(endpoint, community_id=g.community.slug, **kw)


class CommunityEndpoint(Endpoint):

  def get_kwargs(self):
    kwargs = super(CommunityEndpoint, self).get_kwargs()
    kwargs['community_id'] = g.community.slug
    return kwargs


class CommunityTabAction(Action):

  Endpoint = CommunityEndpoint

  def url(self, context=None):
    if self._url:
      return Action.url(self)

    endpoint = self.endpoint
    if endpoint:
      return endpoint
    else:
      return url_for("%s.index" % self.name)

  def is_current(self):
    return g.current_tab == self.name

  
_actions = (
  # Navigation
  NavItem('section', 'communities', title=_l(u'Communities'),
          url=lambda context: url_for_orig('communities.index'),
          condition=lambda ctx: current_user.is_authenticated()),
  # Tabs
  CommunityTabAction('communities:tabs', 'wall', _l(u'Activities')),
  CommunityTabAction('communities:tabs', 'documents', _l(u'Documents'),
                     condition=lambda ctx: g.community.has_documents),
  CommunityTabAction('communities:tabs', 'wiki', _l(u'Wiki'),
                     condition=lambda ctx: g.community.has_wiki),
  CommunityTabAction('communities:tabs', 'forum', _l(u'Conversations'),
                     condition=lambda ctx: g.community.has_forum),
  CommunityTabAction('communities:tabs', 'members', _l(u'Members'),
                     endpoint="communities.members"),
  CommunityTabAction('communities:tabs', 'settings', _l(u'Settings'),
                     icon='cog',
                     condition=lambda ctx: current_user.has_role("admin"),
                     endpoint="communities.settings"),
)


def register_actions(state):
  if not actions.installed(state.app):
    return
  with state.app.app_context():
    actions.register(*_actions)



"""
Not used!
"""

from flask import render_template, request, redirect, g, url_for, flash
from flask_babel import gettext as _
from flask_mail import Message as Email
from sqlalchemy.schema import Column
from sqlalchemy.types import Text
from abilian.core.extensions import db, mail

from .social import social


class Invite(db.Model):
  """Tracks a pending invitation to join the community."""

  __tablename__ = 'invitation'

  email = Column(Text, nullable=False)

  query = db.session.query_property()


@social.route("/invite")
def invite():
  return render_template("social/invite.html")


@social.route("/invite", methods=['POST'])
def invite_post():
  action = request.form.get("action", "cancel")
  if action == 'cancel':
    flash(_(u"Action aborted"), "info")
    return redirect(url_for(".home"))

  emails = request.form.get("emails").split("\n")
  emails = [ email.strip() for email in emails ]

  # FIXME: what do we do with this ?
  message = request.form.get("message", "")  # noqa

  with mail.connect() as conn:
    for email in emails:
      # FIXME: what do we do with this ?
      invite = Invite(sender=g.user, email=email)  # noqa

      #subject = _(u"%s would like to invite you to the %s community") % (g.user.name, "Yaka")
      subject = "%s would like to invite you to the %s community" % (g.user.name, "Yaka")
      msg = Email(subject, recipients=[email], sender=g.user.email)
      params = dict(org_name="Yaka")
      msg.body = render_template('social/mail/invite.txt', **params)
      conn.send(msg)

  flash(_(u"Invitation sent"), "info")
  return redirect(url_for(".home"))



# -*- coding: utf-8 -*-
"""
Document management blueprint.
"""
from __future__ import absolute_import

from flask import g

from abilian.i18n import _l
from abilian.web import nav
from abilian.sbe.apps.communities.blueprint import Blueprint

from ..actions import register_actions

__all__ = ['documents']


documents = Blueprint("documents", __name__,
                      url_prefix="/docs",
                      template_folder="../templates")
route = documents.route
documents.record_once(register_actions)


@documents.url_value_preprocessor
def init_document_values(endpoint, values):
  g.current_tab = 'documents'

  g.breadcrumb.append(
    nav.BreadcrumbItem(label=_l(u'Documents'),
                       url=nav.Endpoint('documents.index',
                                        community_id=g.community.slug)))



from abilian.sbe.testing import BaseTestCase

from ..cmis.renderer import to_xml, Feed
from ..models import Folder, Document

class RendererTestCase(BaseTestCase):

  def test_folder_renderer(self):
    with self.app.test_client():
      folder = Folder(name="tototiti")
      result = to_xml(folder)
      assert "tototiti" in result
      assert "cmis:folder" in result

  def test_document_renderer(self):
    document = Document(name="tototiti")
    result = to_xml(document)
    assert "tototiti" in result
    assert "cmis:document" in result
    assert "<?xml" in result

    result = to_xml(document, no_xml_header=True)
    assert "<?xml" not in result

  def test_feed_renderer(self):
    folder = Folder(title="Toto Titi")
    document = Document(title="tatatutu")
    feed = Feed(folder, [document])
    result = feed.to_xml()

    assert "Toto Titi" in result
    assert "tatatutu" in result



"""
Parses and produces XML documents specified by the standard.
"""

from lxml import objectify, etree
from lxml.builder import ElementMaker


E = ElementMaker(namespace="DAV:")


class Propfind(object):
  def __init__(self, xml=""):
    self.mode = ""
    self.prop_names = []

    if not xml:
      xml = "<D:propfind xmlns:D='DAV:'><D:allprop/></D:propfind>"
    self.parse(xml)

  def parse(self, xml):
    root = objectify.fromstring(xml)

    child = root.getchildren()[0]
    self.mode = child.tag[len("{DAV:}"):]

    if self.mode == "prop":
      for prop in child.getchildren():
        self.prop_names.append(prop.tag)


class MultiStatus(object):
  def __init__(self):
    self.responses = []

  def add_response_for(self, href, obj, property_list):
    response = Response(href, obj, property_list)
    self.responses.append(response)

  def to_string(self):
    return etree.tostring(self.to_xml(), pretty_print=True)

  def to_xml(self):
    xml = E.multistatus()
    for response in self.responses:
      xml.append(response.to_xml())
    return xml


class Response(object):
  def __init__(self, href, obj, property_list):
    self.href = href
    self.property_list = property_list
    self.obj = obj

  def to_xml(self):
    obj = self.obj
    props = E.prop()
    for property_name in self.property_list:
      if property_name == 'creationdate':
        props.append(E.creationdate("1997-12-01T17:42:21-08:00"))
      elif property_name == 'displayname':
        props.append(E.displayname(obj.name))
      elif property_name == 'resourcetype':
        if obj.is_folder:
          props.append(E.resourcetype(E.collection()))
        else:
          props.append(E.resourcetype())

    return E.response(E.href(self.href),
                      E.propstat(props),
                      E.status("HTTP/1.1 200 OK"))



# coding=utf-8
"""
"""
from __future__ import absolute_import

from unittest import TestCase
from datetime import datetime, timedelta
from mock import Mock, patch
from pathlib import Path
from email.parser import FeedParser

import pytest
from flask import url_for
from abilian.core.models.subjects import User

from abilian.sbe.apps.communities.models import MANAGER, MEMBER
from abilian.sbe.testing import BaseTestCase
from abilian.sbe.apps.communities.tests.base import (
  CommunityBaseTestCase, CommunityIndexingTestCase,
)

from ..models import Thread, Post, ThreadClosedError
from ..commands import inject_email
from ..tasks import (
  process, extract_email_destination, build_reply_email_address,
)


class Test(TestCase):
  def test_create_post(self):
    thread = Thread(title="Test thread")
    post = thread.create_post()
    assert post in thread.posts
    assert post.name == u'Test thread'

    thread.title = u'new title'
    assert thread.name == u'new title'
    assert post.name == u'new title'

  def test_closed_property(self):
    thread = Thread(title=u'Test Thread')
    assert thread.closed is False
    thread.closed = True
    assert thread.closed is True
    thread.closed = 0
    assert thread.closed is False
    thread.closed = 1
    assert thread.closed is True
    assert thread.meta['abilian.sbe.forum']['closed'] is True

  def test_thread_closed_guard(self):
    thread = Thread(title=u'Test Thread')
    thread.create_post()
    thread.closed = True

    with pytest.raises(ThreadClosedError):
      thread.create_post()

    p = Post(body_html=u'ok')

    with pytest.raises(ThreadClosedError):
      p.thread = thread

    thread.closed = False
    p.thread = thread
    assert len(thread.posts) == 2

    thread.closed = True
    with pytest.raises(ThreadClosedError):
      del thread.posts[0]

    assert len(thread.posts) == 2

    with pytest.raises(ThreadClosedError):
      # actually thread.posts will be replaced by `[]` and we can't prevent
      # this, but exception has been raised
      thread.posts = []

    thread.closed = False
    thread.posts = [p]
    thread.closed = True
    assert thread.posts == [p]
    assert p.thread is thread
    with pytest.raises(ThreadClosedError):
      p.thread = None


  def test_change_thread_copy_name(self):
    thread = Thread(title=u'thread 1')
    thread2 = Thread(title=u'thread 2')
    post = Post(thread=thread, body_html=u'post content')
    assert post.name == thread.name

    post.thread = thread2
    assert post.name == thread2.name

  def test_task_process_email(self):
    """
      test the process_email function
    """
    marker = u'_____Write above this line to post_____'

    message = get_emailmessage_from_file('reply.email')
    newpost = process(message, marker)
    assert newpost

    message = get_emailmessage_from_file('reply_nocharset_specified.email')
    newpost = process(message, marker)
    assert newpost

    message = get_emailmessage_from_file('reply_no_marker.email')
    self.assertRaises(LookupError, process, message, marker)

    # dubious check
    message = get_emailmessage_from_file('reply_no_textpart.email')
    self.assertRaises(LookupError, process, message, marker)


class IndexingTestCase(CommunityIndexingTestCase):
  def test_thread_indexed(self):
    thread = Thread(title=u'Community 1', community=self.community)
    self.session.add(thread)
    thread_other = Thread(title=u'Community 2: other', community=self.c2)
    self.session.add(thread_other)
    self.session.commit()

    svc = self.svc
    obj_types = (Thread.entity_type,)
    with self.login(self.user_no_community):
      res = svc.search(u'community', object_types=obj_types)
      assert len(res) == 0

    with self.login(self.user):
      res = svc.search(u'community', object_types=obj_types)
      assert len(res) == 1
      hit = res[0]
      assert hit['object_key'] == thread.object_key

    with self.login(self.user_c2):
      res = svc.search(u'community', object_types=obj_types)
      assert len(res) == 1
      hit = res[0]
      assert hit['object_key'] == thread_other.object_key


class NoLoginViewTest(CommunityBaseTestCase):
  """
  Test correct url response, without login or security involved
  """
  def test(self):
    response = self.client.get(
      url_for("forum.index", community_id=self.community.slug))
    self.assert200(response)


class ViewTestCase(CommunityBaseTestCase):
  no_login = False
  SERVICES = ('security',)

  def test_posts_ordering(self):
    thread = Thread(community=self.community, title=u'test ordering')
    self.session.add(thread)
    t1 = datetime(2014, 6, 20, 15, 0, 0)
    p1 = Post(thread=thread, body_html=u'post 1', created_at=t1)
    t2 = datetime(2014, 6, 20, 15, 1, 0)
    p2 = Post(thread=thread, body_html=u'post 2', created_at=t2)
    self.session.flush()
    p1_id, p2_id = p1.id, p2.id
    assert [p.id for p in thread.posts] == [p1_id, p2_id]

    # set post1 created after post2
    t1 = t1 + timedelta(minutes=2)
    p1.created_at = t1
    self.session.flush()
    self.session.expire(thread)  # force thread.posts refreshed from DB
    assert [p.id for p in thread.posts] == [p2_id, p1_id]

  def test_create_thread_and_post(self):
    # activate email reply
    self.app.config['SBE_FORUM_REPLY_BY_MAIL'] = True
    self.app.config['MAIL_ADDRESS_TAG_CHAR'] = u'+'

    # create a new user, add him/her to the current community as a MANAGER
    self.user = User(email=u'user_1@example.com', password='azerty',
                     can_login=True)
    self.session.add(self.user)
    self.community.set_membership(self.user, MANAGER)
    self.session.commit()
    self.client_login(self.user.email, self.user.password)

    mail = self.app.extensions['mail']
    with mail.record_messages() as outbox:
      title = u"Brand new thread"
      content = u"shiny thread message"
      url = url_for("forum.new_thread", community_id=self.community.slug)
      data = dict(title=title, message=content)
      data['__action'] = u"create"
      data['send_by_email'] = u"y"
      response = self.client.post(url, data=data)
      self.assertStatus(response, 302)

      # extract the thread_id from the redirection url in response
      threadid = response.location.rsplit('/', 2)[1]

      # retrieve the new thread, make sur it has the message
      url = url_for("forum.thread", thread_id=threadid,
                    community_id=self.community.slug, title=title)
      response = self.client.get(url)
      self.assertStatus(response, 200)
      self.assertIn(content, response.data.decode("utf-8"))

      # check the email was sent with the new thread
      assert len(outbox) == 1
      assert outbox[0].subject == u'[My Community] Brand new thread'

    # reset the outbox for checking threadpost email
    with mail.record_messages() as outbox:
      content = data['message'] = u"my cherished post"
      del data['title']
      response = self.client.post(url, data=data)
      self.assertStatus(response, 302)

      # retrieve the new thread, make sur it has the message
      url = url_for("forum.thread", thread_id=threadid,
                    community_id=self.community.slug, title=title)
      response = self.client.get(url)
      self.assertStatus(response, 200)
      self.assertIn(content, response.data.decode("utf-8"))

      # check the email was sent with the new threadpost
      assert len(outbox) == 1
      assert unicode(outbox[0].subject) == u'[My Community] Brand new thread'


  def test_create_thread_informative(self):
    """
    Test with 'informative' community. No mail sent, unless user is MANAGER
    """
    assert self.community.type == 'informative'
    # create a new user, add him/her to the current community
    self.user = User(email=u'user_1@example.com', password='azerty',
                     can_login=True)
    self.session.add(self.user)
    self.community.set_membership(self.user, MEMBER)
    self.session.commit()

    title = u"Brand new thread"
    content = u"shiny thread message"
    url = url_for("forum.new_thread", community_id=self.community.slug)
    data = dict(title=title, message=content)
    data['__action'] = u"create"

    mail = self.app.extensions['mail']
    self.client_login(self.user.email, self.user.password)

    with mail.record_messages() as outbox:
      data['send_by_email'] = u"y" # actually should not be in html form
      response = self.client.post(url, data=data)
      self.assertStatus(response, 302)
      assert len(outbox) == 0

    self.community.set_membership(self.user, MANAGER)
    self.session.commit()

    with mail.record_messages() as outbox:
      data['send_by_email'] = u"y" # should be in html form
      response = self.client.post(url, data=data)
      self.assertStatus(response, 302)
      assert len(outbox) == 1

    with mail.record_messages() as outbox:
      del data['send_by_email']
      response = self.client.post(url, data=data)
      self.assertStatus(response, 302)
      assert len(outbox) == 0

    self.client_logout()


def get_string_from_file(filename='notification.email'):
  """
    Load a test email return as string
  """
  filepath = Path(__file__).parent / 'data' / filename
  with filepath.open('rt', encoding='utf-8') as email_file:
    email_string = email_file.read()
  return email_string


def get_emailmessage_from_file(filename='notification.email'):
  """
    Load a mail parse it into a email.message
  """
  email_string = get_string_from_file(filename)
  parser = FeedParser()
  parser.feed(email_string)
  email_message = parser.close()
  return email_message


class CommandsTest(TestCase):
  @patch('fileinput.input')
  @patch('abilian.sbe.apps.forum.commands.process_email')
  def test_parse_forum_email(self, mock_process_email, mock_email):
    """
        No processing is tested only parsing into a email.message
        and verifying inject_email() logic
    """
    # first load a test email returned by the mock_email
    mock_email.return_value = get_string_from_file('notification.email')

    # test the parsing function
    inject_email()

    # assert the email is read
    assert mock_email.called
    # assert a call on the celery task was made implying a message creation
    assert mock_process_email.delay.called


class TasksTest(BaseTestCase):
  def test_build_reply_email_address(self):
    expected_reply_address = 'test+ImZyLzMvNCI.aFudLVO157NrKAGXIj-1Z2kne7E@testcase.app.tld'
    self.app.config['MAIL_ADDRESS_TAG_CHAR'] = u'+'
    post = Mock()
    post.id = 2
    post.thread_id = 3
    member = Mock()
    member.id = 4
    with self.app.test_request_context('/build_reply_email_address', headers=[('Accept-Language', u'fr')]):
      replyto = build_reply_email_address('test', post, member, 'testcase.app.tld')
    self.assertIn(expected_reply_address, replyto)

  def test_extract_mail_destination(self):
    self.app.config['MAIL_ADDRESS_TAG_CHAR'] = u'+'
    self.app.config['MAIL_SENDER'] = u'test@testcase.app.tld'
    test_address = 'test+ImZyLzMvNCI.aFudLVO157NrKAGXIj-1Z2kne7E@testcase.app.tld'
    infos = extract_email_destination(test_address)
    self.assertIn(u'fr', infos[0])  # locale
    self.assertIn(u'3', infos[1])  # thread_id
    self.assertIn(u'4', infos[2])  # post.id



# coding=utf-8
from __future__ import absolute_import

import re
from urllib import quote_plus

from mock import MagicMock, patch
from markdown import Markdown
from flask import url_for, g
from abilian.core.models.subjects import User

from abilian.sbe.apps.communities.tests.base import (
  CommunityBaseTestCase, CommunityIndexingTestCase
)
from abilian.sbe.apps.wiki.markup import SBEWikiLinkExtension
from .models import WikiPage
from . import views


def test_wikilink_extension():
  text = u'/*€('
  wikilink = u'[[' + text + ']]'

  def build_url(label, base, end):
    return u'/?title=' + quote_plus(label.encode('utf-8')) + end

  extension = SBEWikiLinkExtension([('build_url', build_url)])
  ctx = {}
  ctx['extensions'] = [extension, 'toc']
  ctx['output_format'] = 'html5'
  md = Markdown(**ctx)

  page_exists_mock = MagicMock(return_value=True)
  with patch('abilian.sbe.apps.wiki.forms.page_exists', page_exists_mock):
    result = md.convert(wikilink)

  qtext = unicode(quote_plus(text.encode('utf-8')))
  expected = u'<p><a class="wikilink" href="/?title={href}/">{text}</a></p>'
  expected = expected.format(href=qtext, text=text)
  assert expected == result

  # make sur the class is 'wikilink new' if page exists'
  page_exists_mock = MagicMock(return_value=False)
  with patch('abilian.sbe.apps.wiki.forms.page_exists', page_exists_mock):
    result = md.convert(wikilink)

  qtext = unicode(quote_plus(text.encode('utf-8')))
  expected = u'<p><a class="wikilink new" href="/?title={href}/">{text}</a></p>'
  expected = expected.format(href=qtext, text=text)
  assert expected == result


class WikiBaseTestCase(CommunityBaseTestCase):
  def setUp(self):
    super(WikiBaseTestCase, self).setUp()
    g.user = User.query.all()[0]


class TestModels(WikiBaseTestCase):
  def test_create_page(self):
    page = WikiPage(title=u"Some page", body_src=u'abc')
    self.assertEquals(page.title, u'Some page')
    self.assertEquals(page.name, u'Some page')
    self.assertEquals(page.body_src, 'abc')
    self.assertEquals(page.body_html, '<p>abc</p>')
    self.assertEquals(len(page.revisions), 1)

    revision = page.revisions[0]
    self.assertEquals(revision.number, 0)
    self.assertEquals(revision.author, g.user)

  def test_rename_page(self):
    page = WikiPage(title=u"Some page", body_src=u'abc')
    self.assertEquals(page.title, u'Some page')
    self.assertEquals(page.name, u'Some page')
    page.title = u'Title Renamed'
    self.assertEquals(page.title, u'Title Renamed')
    self.assertEquals(page.name, u'Title Renamed')

    page.name = u'Name'
    self.assertEquals(page.title, u'Name')
    self.assertEquals(page.name, u'Name')

  def test_create_revision(self):
    page = WikiPage('abc')
    page.create_revision("def", "page updated")

    self.assertEquals(len(page.revisions), 2)
    last_revision = page.revisions[1]
    self.assertEquals(last_revision.number, 1)
    self.assertEquals(last_revision.author, g.user)


class TestIndexing(CommunityIndexingTestCase):

  def setUp(self):
    super(TestIndexing, self).setUp()
    g.user = User.query.all()[0]

  def test_wiki_indexed(self):
    page = WikiPage(title=u'Community 1', community=self.community)
    self.session.add(page)
    page_other = WikiPage(title=u'Community 2: other', community=self.c2)
    self.session.add(page_other)
    self.session.commit()

    svc = self.svc
    obj_types = (WikiPage.entity_type,)
    with self.login(self.user_no_community):
      res = svc.search(u'community', object_types=obj_types)
      self.assertEquals(len(res), 0)

    with self.login(self.user):
      res = svc.search(u'community', object_types=obj_types)
      self.assertEquals(len(res), 1)
      hit = res[0]
      self.assertEquals(hit['object_key'], page.object_key)

    with self.login(self.user_c2):
      res = svc.search(u'community', object_types=obj_types)
      self.assertEquals(len(res), 1)
      hit = res[0]
      self.assertEquals(hit['object_key'], page_other.object_key)


class TestsViews(WikiBaseTestCase):
  def test_home(self):
    response = self.client.get(
      url_for("wiki.index", community_id=self.community.slug))
    self.assertStatus(response, 302)

    response = self.client.get(
      url_for("wiki.page", title='Home', community_id=self.community.slug))
    assert response.status_code == 200

  def test_create_page_initial_form(self):
    g.community = self.community
    view = views.PageCreate()
    view.prepare_args([], {})
    form = view.form
    assert form['last_revision_id'].data is None

  def test_create_page(self):
    title = 'Some page name'
    url = url_for("wiki.page_new", community_id=self.community.slug)
    url += '?title=Some+page+name'
    response = self.client.get(url)
    assert response.status_code == 200
    # make sure the title is filled when comming from wikilink
    self.assertIn('value="Some page name"', response.data)

    title = 'Some page'
    body = "LuuP3jai"
    url = url_for("wiki.page_new", community_id=self.community.slug)
    data = dict(title=title, body_src=body, __action="create")
    response = self.client.post(url, data=data)
    assert response.status_code == 302

    url = url_for("wiki.page", community_id=self.community.slug, title=title)
    response = self.client.get(url)
    assert response.status_code == 200
    self.assertIn(body, response.data)

    # edit
    url = url_for("wiki.page_edit",
                  community_id=self.community.slug,
                  title=title)
    response = self.client.get(url)
    assert response.status_code == 200

    # Slightly hackish way to get the page_id
    line = [l for l in response.data.split("\n") if 'name="page_id"' in l][0]
    m = re.search('value="(.*?)"', line)
    page_id = int(m.group(1))

    url = url_for("wiki.page_changes", community_id=self.community.slug,
                  title=title)
    response = self.client.get(url)
    assert response.status_code == 200

    url = url_for("wiki.page_source", community_id=self.community.slug,
                  title=title)
    response = self.client.get(url)
    assert response.status_code == 200

    url = url_for("wiki.page_edit", community_id=self.community.slug)
    data = dict(title=title, page_id=page_id, body_src="abc def")
    data['__action'] = 'edit'
    response = self.client.post(url, data=data)
    assert response.status_code == 302

    url = url_for("wiki.page_compare", rev0="on", rev1="on",
                  community_id=self.community.slug, title=title)
    response = self.client.get(url)
    assert response.status_code == 200



# coding=utf-8
"""
"""
from __future__ import absolute_import
from unittest import skip

import mock
import sqlalchemy as sa
from flask import url_for

from abilian.core.entities import Entity
from abilian.core.models.subjects import User
from abilian.services.security import Admin
from abilian.sbe.apps.documents.models import Folder
from abilian.sbe.testing import BaseTestCase

from ..models import (
  Community, community_content, CommunityIdColumn,
  MEMBER,
)
from .. import views, signals
from .base import CommunityIndexingTestCase as BaseIndexingTestCase


class CommunityUnitTestCase(BaseTestCase):

  def test_instanciation(self):
    community = Community(name=u"My Community")
    assert isinstance(community.folder, Folder)
    #assert isinstance(community.group, Group)

  def test_default_view_kw(self):
    # test exceptions are handled if passed an object with 'community' attribute
    # and no community_id in kwargs. and ValueError is properly raised
    dummy = type('Dummy', (object,), {'community': None})()
    with self.assertRaises(ValueError) as cm:
      views.default_view_kw({}, dummy, 'dummy', 1)

    assert cm.exception.message == 'Cannot find community_id value'


class CommunityDbTestCase(BaseTestCase):

  def setUp(self):
    super(CommunityDbTestCase, self).setUp()
    self.community = Community(name=u"My Community")
    self.db.session.add(self.community)
    self.db.session.commit()

  def test_default_url(self):
    assert self.app.default_view.url_for(self.community) \
           == '/communities/my-community/'

  def test_can_recreate_with_same_name(self):
    name = self.community.name
    self.db.session.delete(self.community)
    self.db.session.commit()
    self.community = Community(name=name)
    self.db.session.add(self.community)
    # if community.folder was not deleted, this will raise IntegrityError. Test
    # passes if no exceptions is raised
    self.db.session.commit()

  def test_rename(self):
    NEW_NAME = u"My new name"
    self.community.rename(NEW_NAME)
    assert self.community.name == NEW_NAME
    assert self.community.folder.name == NEW_NAME

  def test_auto_slug(self):
    assert self.community.slug == u'my-community'

  def test_membership(self):
    user = User.query.first()
    l = self.community.memberships
    assert l == []

    # setup signals testers with mocks.
    when_set = mock.MagicMock()
    when_set.mock_add_spec(['__name__']) # required for signals
    signals.membership_set.connect(when_set)
    when_removed = mock.MagicMock()
    when_removed.mock_add_spec(['__name__'])
    signals.membership_removed.connect(when_removed)

    # invalid role
    with self.assertRaises(ValueError):
      self.community.set_membership(user, 'dummy role name')

    assert not when_set.called
    assert not when_removed.called

    # simple member
    self.community.set_membership(user, "member")
    self.db.session.commit()

    l = self.community.memberships
    assert len(l) == 1
    assert l[0].user == user
    assert l[0].role is MEMBER

    when_set.assert_called_once_with(self.community,
                                     is_new=True,
                                     membership=l[0])
    assert not when_removed.called
    when_set.reset_mock()

    assert self.community.get_role(user) is MEMBER

    assert self.community.get_memberships() == [l[0]]
    assert self.community.get_memberships('member') == [l[0]]
    assert self.community.get_memberships('manager') == []

    # change user role
    self.community.set_membership(user, "manager")
    self.db.session.commit()

    l = self.community.memberships
    assert len(l) == 1
    assert l[0].user == user
    assert l[0].role == "manager"

    assert self.community.get_role(user) == "manager"

    when_set.assert_called_once_with(self.community,
                                     is_new=False,
                                     membership=l[0])
    assert not when_removed.called
    when_set.reset_mock()

    # remove user
    membership = l[0]
    self.community.remove_membership(user)
    self.db.session.commit()

    l = self.community.memberships
    assert l == []

    assert not when_set.called
    when_removed.assert_called_once_with(self.community,
                                         membership=membership)

  def test_folder_roles(self):
    user = User.query.first()
    folder = self.community.folder
    self.community.set_membership(user, "member")
    self.db.session.commit()
    security = self.app.services['security']

    assert security.get_roles(user, folder) == ["reader"]

    # this tests a bug, where local roles whould disappear when setting
    # membership twice
    self.community.set_membership(user, "member")
    assert security.get_roles(user, folder) == ["reader"]


  def test_community_content_decorator(self):
    @community_content
    class CommunityContent(Entity):
      community_id = CommunityIdColumn()
      community = sa.orm.relation(Community, foreign_keys=[community_id])

    sa.orm.configure_mappers()
    conn = self.session.connection()
    for table in sa.inspect(CommunityContent).tables:
      if not table.exists(conn):
        table.create(conn)

    cc = CommunityContent(name=u'my content', community=self.community)
    self.session.add(cc)
    self.session.flush()
    assert hasattr(cc, 'community_slug')
    assert cc.community_slug == 'my-community'
    assert cc.slug == u'my-content'

    index_to = dict(CommunityContent.__indexation_args__['index_to'])
    assert 'community_slug' in index_to


class CommunityIndexingTestCase(BaseIndexingTestCase):

  def test_community_indexed(self):
    svc = self.svc
    obj_types = (Community.entity_type,)
    with self.login(self.user_no_community):
      res = svc.search(u'community', object_types=obj_types)
      assert len(res) == 0

    with self.login(self.user):
      res = svc.search(u'community', object_types=obj_types)
      assert len(res) == 1
      hit = res[0]
      assert hit['object_key'] == self.community.object_key

    with self.login(self.user_c2):
      res = svc.search(u'community', object_types=obj_types)
      assert len(res) == 1
      hit = res[0]
      assert hit['object_key'] == self.c2.object_key

  def test_default_view_kw_with_hit(self):
    with self.login(self.user):
      hit = self.svc.search(u'community',
                            object_types=(Community.entity_type,))[0]
      kw = views.default_view_kw({}, hit, hit['object_type'], hit['id'])

    assert kw == {'community_id': self.community.slug}


class CommunityWebTestCase(BaseIndexingTestCase):

  def test_index(self):
    with self.client_login(self.user.email, 'azerty'):
      response = self.client.get(url_for("communities.index"))
    self.assert_200(response)

  def test_community_home(self):
    url = self.app.default_view.url_for(self.community)
    user = self.user.email
    user_c2 = self.user_c2.email
    with self.client_login(user_c2, 'azerty'):
      response = self.client.get(url)
      assert response.status_code == 403

    with self.client_login(user, 'azerty'):
      response = self.client.get(url)
      assert response.status_code == 302
      expected_url = url_for("wall.index", community_id=self.community.slug,
                             _external=True)
      assert response.headers['Location'] == expected_url

  def test_community_settings(self):
    url = url_for('communities.settings', community_id=self.community.slug)
    with self.client_login(self.user.email, 'azerty'):
      response = self.client.get(url)
      assert response.status_code == 403

      self.app.services['security'].grant_role(self.user, Admin)
      response = self.client.get(url)
      self.assert_200(response)

      data = {'__action': u'edit',
              'name': u'edited community',
              'description': u'my community',
              'type': 'participative',
      }
      response = self.client.post(url, data=data)
      assert response.status_code == 302
      self.assertEquals(
        response.headers['Location'],
        u'http://localhost/communities/{}/'.format(self.community.slug))

      community = Community.query.get(self.community.id)
      assert community.name == u'edited community'

  def test_new(self):
    with self.client_login(self.user.email, 'azerty'):
      response = self.client.get(url_for("communities.new"))
      assert response.status_code == 403

      self.app.services['security'].grant_role(self.user, Admin)
      response = self.client.get(url_for("communities.new"))
      assert response.status_code == 200

  def test_members(self):
    with self.client_login(self.user.email, 'azerty'):
      url = url_for("communities.members", community_id=self.community.slug)
      response = self.client.get(url)
      self.assert_200(response)

      # test add user
      data = {'action': 'add-user-role', 'user': self.user_c2.id}
      response = self.client.post(url, data=data)
      assert response.status_code == 403

      self.app.services['security'].grant_role(self.user, Admin)
      data = {
        'action': 'add-user-role',
        'user': self.user_c2.id,
        'role': 'member',
      }
      response = self.client.post(url, data=data)
      assert response.status_code == 302
      response.headers['Location'] == 'http://localhost' + url

      membership = [m for m in self.community.memberships
                    if m.user == self.user_c2][0]
      assert membership.role == 'member'

      data['action'] = 'set-user-role'
      data['role'] = 'manager'
      response = self.client.post(url, data=data)
      assert response.status_code == 302
      assert response.headers['Location'] == 'http://localhost' + url
      self.session.expire(membership)
      assert membership.role == 'manager'

      # Community.query.session is not self.db.session, but web app session.
      community = Community.query.get(self.community.id)
      assert self.user_c2 in community.members

      # test delete
      data = {
        'action': 'delete',
        'user': self.user_c2.id,
        'membership': [m.id for m in community.memberships
                       if m.user == self.user_c2][0],
      }
      response = self.client.post(url, data=data)
      assert response.status_code == 302
      self.assertEquals(
        response.headers['Location'],
        u'http://localhost/communities/{}/members'.format(self.community.slug))

      assert self.user_c2 not in community.members



#!./bin/python

import subprocess
import urllib
import time

# Some random number
BIND = '0.0.0.0'
PORT = 4034
HOME = "http://{}:{}/".format(BIND, PORT)

subprocess.call(['./bin/pip', 'install', 'gunicorn'])

p = subprocess.Popen(["./bin/gunicorn",
                      'wsgi:app',
                      '-b', '{}:{}'.format(BIND, PORT)])

try:
  # Just in case
  time.sleep(5)
  page = urllib.urlopen(HOME).read()
  assert "Welcome to Abilian" in page
finally:
  p.terminate()
  p.wait()



# coding=utf-8
"""
Forms for the Wiki module.
"""
from __future__ import absolute_import

from flask import g
from wtforms import TextAreaField, StringField, HiddenField
from wtforms.validators import ValidationError, required

from abilian.i18n import _, _l
from abilian.web.forms import Form
from abilian.web.forms.widgets import TextArea
from abilian.web.forms.validators import flaghidden
from abilian.web.forms.filters import strip

from .models import WikiPage


def clean_up(src):
  """
  form filter
  """
  src = src.replace("\r", "")
  return src


def int_or_none(val):
  try:
    return int(val)
  except (TypeError, ValueError):
    return None


class WikiPageForm(Form):
  title = StringField(label=_l(u"Title"), filters=(strip,),
                      validators=[required()])
  body_src = TextAreaField(
      label=_l("Body"),
      filters=(strip, clean_up),
      validators=[required()],
      widget=TextArea(rows=10, resizeable='vertical'),
  )

  message = StringField(label=_l("Commit message"))
  page_id = HiddenField(filters=(int_or_none,), validators=[flaghidden()])
  last_revision_id = HiddenField(filters=(int_or_none,),
                                 validators=[flaghidden()])

  def validate_title(self, field):
    title = field.data
    if title != field.object_data and page_exists(title):
      raise ValidationError(
          _(u"A page with this name already exists. Please use another name.")
      )

  def validate_last_revision_id(self, field):
    val = field.data
    current = field.object_data

    if val is None or current is None:
      return

    if val != current:
      raise ValidationError(_(u'this page has been edited since'))


def page_exists(title):
  title = title.strip()
  return WikiPage.query\
      .filter(WikiPage.community_id == g.community.id,
              WikiPage.title == title)\
      .count() > 0


# Not used yet
class CommentForm(Form):
  message = TextAreaField(label=_l("Message"), validators=[required()])
  page_id = HiddenField()



from __future__ import print_function

import uuid
import os.path

from lxml.etree import XMLSyntaxError
from flask import Blueprint, request, current_app
from werkzeug.exceptions import NotFound, Forbidden
from werkzeug.wrappers import BaseResponse as Response
from werkzeug.datastructures import Headers

from abilian.core.extensions import db
from .. import repository
from .constants import *
from .xml import Propfind, MultiStatus


webdav = Blueprint("webdav", __name__, url_prefix="/webdav")
route = webdav.route

__all__ = ['webdav']


#
# Utils
#

# TODO: real logging
class Logger(object):
  def debug(self, msg):
    print(msg)

log = Logger()


# XXX: temporary debug info.
@webdav.before_request
def log_request():
  litmus_msg = request.headers.get('X-Litmus')
  if litmus_msg:
    print('')
    print(78 * '#')
    print(litmus_msg)
    print('{} on {}'.format(request.method, request.path))


@webdav.before_request
def only_admin():
  if not current_app.services('security').has_role(user, "admin"):
    raise Forbidden()


@webdav.after_request
def log_response(response):
  print('Response: {}'.format(response))
  return response


def normpath(path):
  path = os.path.normpath(path)
  if not path.startswith("/"):
    path = "/" + path
  return path


def split_path(path):
  path = normpath(path)
  return os.path.dirname(path), os.path.basename(path)


def get_object(path):
  obj = repository.get_object_by_path(path)
  if obj is None:
    raise NotFound()
  return obj


@webdav.before_app_request
def create_root_folder():
  # TODO: create root folder on repository startup instead.
  #assert repository.root_folder
  pass


#
# HTTP endpoints
#
@route('/', methods=['OPTIONS'], defaults={'path': '/'})
@route('/<path:path>', methods=['OPTIONS'])
def options(path):
  headers = {'Content-Type': 'text/plain',
             'Content-Length': '0',
             'DAV': '1,2',
             'MS-Author-Via': 'DAV',
             'Allow': OPTIONS}
  print("Returning", headers)
  return "", HTTP_OK, headers


@route('/', defaults={'path': '/'})
@route('/<path:path>')
def get(path):
  path = normpath(path)

  obj = get_object(path)

  file_name = obj.file_name.encode('utf8')
  headers = {'Content-Type': obj.content_type,
             'Content-Disposition': 'attachment;filename=%s' % file_name}
  return obj.content, HTTP_OK, headers


@route('/<path:path>', methods=['MKCOL'])
def mkcol(path):
  path = normpath(path)
  parent_path, name = split_path(path)

  if request.data:
    return "Request body must be empty.", 415, {}

  obj = repository.get_object_by_path(path)
  if obj is not None:
    return "Objet already exists.", HTTP_METHOD_NOT_ALLOWED, {}

  parent_folder = repository.get_folder_by_path(parent_path)
  if parent_folder is None:
    return "Parent collection doesn't exist.", HTTP_CONFLICT, {}

  new_folder = parent_folder.create_subfolder(name=name)
  db.session.commit()
  return "", HTTP_CREATED, {}


@route('/<path:path>', methods=['PUT'])
def put(path):
  path = normpath(path)
  parent_path, name = split_path(path)

  status = HTTP_CREATED
  obj = repository.get_object_by_path(path)
  if obj is not None:
    if not obj.is_document:
      return "", HTTP_METHOD_NOT_ALLOWED, {}
    else:
      status = HTTP_NO_CONTENT
  else:
    parent_folder = repository.get_folder_by_path(parent_path)
    if parent_folder is None:
      return "", HTTP_CONFLICT, {}
    obj = parent_folder.create_document(name=name)

  obj.content = request.data
  obj.content_type = request.content_type

  db.session.commit()
  return "", status, {}


@route('/<path:path>', methods=['DELETE'])
def delete(path):
  path = normpath(path)

  obj = get_object(path)

  db.session.delete(obj)
  db.session.commit()
  return "", HTTP_NO_CONTENT, {}


@route("/<path:path>", methods=['COPY', 'MOVE'])
def copy_or_move(path):
  path = normpath(path)
  dest = request.headers.get('destination')
  dest_path = normpath(dest[len(request.url_root + "webdav"):])
  dest_parent_path, dest_name = split_path(dest_path)
  overwrite = request.headers.get('overwrite')

  orig_obj = get_object(path)

  status = HTTP_CREATED
  dest_obj = repository.get_object_by_path(dest_path)
  if dest_obj:
    if overwrite == "F":
      return "", HTTP_PRECONDITION_FAILED, {}
    else:
      repository.delete_object(dest_obj)
      db.session.flush()
      status = HTTP_NO_CONTENT

  dest_folder = repository.get_folder_by_path(dest_parent_path)
  if dest_folder is None:
    return "", HTTP_CONFLICT, {}

  if request.method == 'COPY':
    repository.copy_object(orig_obj, dest_folder, dest_name)
  else:
    if dest_folder == orig_obj.parent:
      repository.rename_object(orig_obj, dest_name)
    else:
      repository.move_object(orig_obj, dest_folder, dest_name)

  db.session.commit()
  return "", status, {}


@route('/', defaults={'path': '/'}, methods=['PROPFIND'])
@route('/<path:path>', methods=['PROPFIND'])
def propfind(path):
  path = normpath(path)
  depth = request.headers.get('depth', "1")

  print(request.headers)
  print(request.data)

  try:
    propfind = Propfind(request.data)
  except XMLSyntaxError:
    return "Malformed XML document.", HTTP_BAD_REQUEST, {}

  obj = get_object(path)

  m = MultiStatus()
  m.add_response_for(request.url, obj, DAV_PROPS)

  if depth == "1" and obj.is_folder:
    for child in obj.children:
      m.add_response_for(request.url + "/" + child.name, child, DAV_PROPS)

  print(m.to_string())

  headers = {'Content-Type': 'text/xml'}
  return m.to_string(), HTTP_MULTI_STATUS, headers


@route('/<path:path>', methods=['LOCK'])
def lock(path):
  path = normpath(path)
  obj = get_object(path)
  token = str(uuid.uuid1())

  if repository.is_locked(obj):
    if not repository.can_unlock(obj):
      return "", 423, {}
    else:
      headers = {"Lock-Token": "urn:uuid:" + token}
      return "TODO", HTTP_OK, headers

  token = repository.lock(obj)


  xml = '''<?xml version="1.0" encoding="utf-8" ?>
<D:prop xmlns:D="DAV:">
    <D:lockdiscovery>
        <D:activelock>
            <D:lockscope><D:exclusive/></D:lockscope>
            <D:locktype><D:write/></D:locktype>
            <D:depth>0</D:depth>
            <D:timeout>Second-179</D:timeout>
            <D:owner>flora</D:owner>
            <D:locktoken>
                <D:href>opaquelocktoken:%s</D:href>
            </D:locktoken>
        </D:activelock>
    </D:lockdiscovery>
</D:prop>''' % token

  hlist = [('Content-Type', 'text/xml'),
           ('Lock-Token', '<urn:uuid:%s>' % token)]

  return Response(xml, headers=Headers.linked(hlist))  #, status ='423 Locked'


  """
    public Response lock(@Context UriInfo uriInfo) throws Exception {
        String token = null;
        Prop prop = null;
        if (backend.isLocked(doc.getRef())) {
            if (!backend.canUnlock(doc.getRef())) {
                return Response.status(423).build();
            } else {
                token = backend.getCheckoutUser(doc.getRef());
                prop = new Prop(getLockDiscovery(doc, uriInfo));
                return Response.ok().entity(prop).header("Lock-Token",
                        "urn:uuid:" + token).build();
            }
        }

        token = backend.lock(doc.getRef());
        if (READONLY_TOKEN.equals(token)) {
            return Response.status(423).build();
        } else if (StringUtils.isEmpty(token)) {
            return Response.status(400).build();
        }

        prop = new Prop(getLockDiscovery(doc, uriInfo));

        backend.saveChanges();
        return Response.ok().entity(prop).header("Lock-Token",
                "urn:uuid:" + token).build();
    }

  """

@route('/<path:path>', methods=['UNLOCK'])
def unlock(path):
  path = normpath(path)
  obj = get_object(path)

  if repository.is_locked(obj):
    if not repository.can_unlock(obj):
      return "", 423, {}
    else:
      repository.unlock(obj)
      db.session.commit()
      return "", HTTP_NO_CONTENT, {}

  return "", HTTP_NO_CONTENT, {}

  """
        if (backend.isLocked(doc.getRef())) {
            if (!backend.canUnlock(doc.getRef())) {
                return Response.status(423).build();
            } else {
                backend.unlock(doc.getRef());
                backend.saveChanges();
                return Response.status(HTTP_NO_CONTENT).build();
            }
        } else {
            // TODO: return an error
            return Response.status(HTTP_NO_CONTENT).build();
        }


  """



# coding=utf-8
"""
"""
from __future__ import absolute_import

from cStringIO import StringIO
from zipfile import ZipFile
from os.path import join, dirname
from itertools import count
import unittest

from werkzeug.datastructures import FileStorage
from flask import g, get_flashed_messages
from abilian.web.util import url_for
from abilian.sbe.apps.communities.presenters import CommunityPresenter
from abilian.sbe.apps.communities.models import WRITER
from abilian.sbe.apps.communities.tests.base import (
  CommunityBaseTestCase, CommunityIndexingTestCase,
)
from abilian.sbe.apps.documents.models import (
  Folder, db, Document,
  PathAndSecurityIndexable,
)
from abilian.sbe.apps.documents.views import util as view_util


class BaseTests(CommunityBaseTestCase):
  init_data = True
  no_login = True

  def setUp(self):
    super(BaseTests, self).setUp()

    self.root_folder = Folder(title=u"root")
    db.session.add(self.root_folder)
    db.session.flush()

  @staticmethod
  def uid_from_url(url):
    return int(url.split("/")[-1])

  @staticmethod
  def open_file(filename):
    path = join(dirname(__file__), "data", "dummy_files", filename)
    return open(path, 'rb')


class TestBlobs(BaseTests):
  def test_document(self):
    root = Folder(title=u"root")
    doc = Document(parent=root, title=u"test")
    data = self.open_file("onepage.pdf").read()
    doc.set_content(data, "application/pdf")
    self.session.add(doc)
    self.session.commit()

    # coverage
    self.app.config['ANTIVIRUS_CHECK_REQUIRED'] = True
    doc.ensure_antivirus_scheduled()

  def test_antivirus_properties(self):
    root = Folder(title=u"root")
    doc = Document(parent=root, title=u"test")
    doc.set_content('content', 'text/plain')
    appcfg = self.app.config
    meta = doc.content_blob.meta

    # 1: not check required ################
    # no data in meta
    appcfg['ANTIVIRUS_CHECK_REQUIRED'] = False
    assert doc.antivirus_scanned is False
    assert doc.antivirus_status is None
    assert doc.antivirus_required is False
    assert doc.antivirus_ok is True

    # antivirus was run, but no result
    meta['antivirus'] = None
    assert doc.antivirus_scanned is True
    assert doc.antivirus_status is None
    assert doc.antivirus_required is False
    assert doc.antivirus_ok is True

    # virus detected
    meta['antivirus'] = False
    assert doc.antivirus_scanned is True
    assert doc.antivirus_status is False
    assert doc.antivirus_required is False
    assert doc.antivirus_ok is False

    # virus free
    meta['antivirus'] = True
    assert doc.antivirus_scanned is True
    assert doc.antivirus_status is True
    assert doc.antivirus_required is False
    assert doc.antivirus_ok is True

    # 2: check required ##################
    # no data in meta
    del meta['antivirus']
    appcfg['ANTIVIRUS_CHECK_REQUIRED'] = True
    assert doc.antivirus_scanned is False
    assert doc.antivirus_status is None
    assert doc.antivirus_required is True
    assert doc.antivirus_ok is False

    # antivirus was run, but no result
    meta['antivirus'] = None
    assert doc.antivirus_scanned is True
    assert doc.antivirus_status is None
    assert doc.antivirus_required is True
    assert doc.antivirus_ok is False

    # virus detected
    meta['antivirus'] = False
    assert doc.antivirus_scanned is True
    assert doc.antivirus_status is False
    assert doc.antivirus_required is False
    assert doc.antivirus_ok is False

    # virus free
    meta['antivirus'] = True
    assert doc.antivirus_scanned is True
    assert doc.antivirus_status is True
    assert doc.antivirus_required is False
    assert doc.antivirus_ok is True


class IndexingTestCase(CommunityIndexingTestCase):
  def test_folder_indexed(self):
    folder = Folder(title=u'Folder 1', parent=self.community.folder)
    self.session.add(folder)
    folder_other = Folder(title=u'Folder 2: other', parent=self.c2.folder)
    self.session.add(folder_other)
    self.session.commit()

    svc = self.svc
    obj_types = (Folder.entity_type,)
    with self.login(self.user_no_community):
      res = svc.search(u'folder', object_types=obj_types)
      assert len(res) == 0

    with self.login(self.user):
      res = svc.search(u'folder', object_types=obj_types)
      assert len(res) == 1
      hit = res[0]
      assert hit['object_key'] == folder.object_key
      assert hit['community_slug'] == self.community.slug

    with self.login(self.user_c2):
      res = svc.search(u'folder', object_types=obj_types)
      assert len(res) == 1
      hit = res[0]
      assert hit['object_key'] == folder_other.object_key
      assert hit['community_slug'] == self.c2.slug


class TestViews(CommunityIndexingTestCase, BaseTests):
  def setUp(self):
    super(TestViews, self).setUp()
    self.community.type = 'participative'
    self.community.set_membership(self.user, WRITER)
    self.session.commit()
    self.folder = self.community.folder

  def test_util_create(self):
    with self.login(self.user):
      g.user = self.user
      g.community = CommunityPresenter(self.community)
      name = u'document'
      fs = FileStorage(StringIO(u'content'), filename=name,
                       content_type='text/plain')
      doc = view_util.create_document(self.folder, fs)
      self.session.flush()
      assert doc.parent == self.folder
      assert doc.name == name

      # test upload with same name: should be renamed
      fs = FileStorage(StringIO(u'content'), filename=name,
                       content_type='text/plain')
      doc2 = view_util.create_document(self.folder, fs)
      self.session.flush()
      assert doc2.parent == self.folder
      assert len(self.folder.children) == 2
      assert doc2.name == name + u'-1'
      messages = get_flashed_messages()
      assert len(messages) == 1

  def test_home(self):
    with self.client_login(self.user.email, password='azerty'):
      response = self.get(url_for('documents.index',
                                  community_id=self.community.slug))
      self.assert_status(response, 302)
      self.assertEquals(
        response.headers['Location'],
        u'http://localhost/communities/{}/docs/folder/{}'
        u''.format(self.community.slug, self.folder.id),
      )

  def _test_upload(self, title, content_type, test_preview=True,
                   assert_preview_available=True):
    data = {
      'file': (self.open_file(title), title, content_type),
      'action': 'upload',
    }

    folder = self.community.folder
    url = url_for("documents.folder_post",
                  community_id=self.community.slug,
                  folder_id=folder.id)
    response = self.client.post(url, data=data)
    self.assert_status(response, 302)

    doc = folder.children[0]
    assert doc.title == title
    url = url_for("documents.document_view",
                  community_id=self.community.slug, doc_id=doc.id)
    response = self.get(url)
    self.assert_200(response)

    url = url_for("documents.document_download",
                  community_id=self.community.slug,
                  doc_id=doc.id)
    response = self.get(url)
    self.assert_200(response)
    assert response.headers['Content-Type'] == content_type

    content = self.open_file(title).read()
    assert response.data == content

    if test_preview:
      url = url_for("documents.document_preview_image",
                    community_id=self.community.slug,
                    doc_id=doc.id, size=500)
      response = self.get(url)
      if assert_preview_available:
        self.assert_200(response)
        assert response.headers['Content-Type'] == 'image/jpeg'
      else:
        # redirect to 'missing image'
        self.assert_302(response)
        assert response.headers['Cache-Control'] == 'no-cache'

    url = url_for("documents.document_delete",
                  community_id=self.community.slug, doc_id=doc.id)
    response = self.client.post(url)
    self.assert_302(response)

    url = url_for("documents.document_view",
                  community_id=self.community.slug, doc_id=doc.id)
    response = self.get(url)
    self.assert_404(response)

  def test_text_upload(self):
    NAME = 'wikipedia-fr.txt'
    with self.client_login(self.user.email, password='azerty'):
      self._test_upload(NAME, "text/plain", test_preview=False)

  def test_pdf_upload(self):
    NAME = u"onepage.pdf"
    with self.client_login(self.user.email, password='azerty'):
      self._test_upload(NAME, "application/pdf")

  def test_image_upload(self):
    NAME = u"picture.jpg"
    with self.client_login(self.user.email, password='azerty'):
      self._test_upload(NAME, "image/jpeg")

  def test_binary_upload(self):
    NAME = u"random.bin"
    with self.client_login(self.user.email, password='azerty'):
      self._test_upload(NAME, "application/octet-stream",
                        assert_preview_available=False)

  def test_explore_archive(self):
    from abilian.sbe.apps.documents.views.folders import explore_archive
    fd = self.open_file('content.zip')
    result = [(u'/'.join(path), f)
              for path, f in explore_archive(fd, u'content.zip')]
    assert result == [(u'', fd)]

    fd = self.open_file('content.zip')
    archive_content = explore_archive(fd, u'content.zip', uncompress=True)
    result = {u'/'.join(path) + u'/' + f.filename
              for path, f in archive_content}
    assert result == {
      u'existing-doc/file.txt',
      u'existing-doc/subfolder_in_renamed/doc.txt',
      u'folder 1/doc.txt',
      u'folder 1/dos cp437: é.txt',
      u'folder 1/osx: utf-8: é.txt'
    }

  def test_zip_upload_uncompress(self):
    folder = Folder(title=u'folder 1', parent=self.community.folder)
    self.session.add(folder)
    self.session.flush()
    folder = self.community.folder
    files = []
    files.append((StringIO('A document'), u'existing-doc', 'text/plain'))
    files.append((self.open_file('content.zip'),
                  u'content.zip',
                  'application/zip'))
    data = {
      'file': files,
      'action': 'upload',
      'uncompress_files': True
    }
    url = url_for("documents.folder_post",
                  community_id=self.community.slug,
                  folder_id=folder.id)
    with self.client_login(self.user.email, password='azerty'):
      response = self.client.post(url, data=data)

    expected = {u'existing-doc', u'folder 1', u'existing-doc-1'}
    self.assert_status(response, 302)
    assert expected == {f.title for f in folder.children}
    expected = {u'folder 1', u'existing-doc-1'}
    assert expected == {f.title for f in folder.subfolders}

  def test_zip(self):
    with self.client_login(self.user.email, password='azerty'):
      title = u"onepage.pdf"
      content_type = "application/pdf"
      data = {
        'file': (self.open_file(title), title, content_type),
        'action': 'upload',
      }
      folder = self.community.folder
      url = url_for("documents.folder_post",
                    community_id=self.community.slug, folder_id=folder.id)
      response = self.client.post(url, data=data)
      self.assert_302(response)

      doc = folder.children[0]
      data = {'action': 'download',
              'object-selected': ["cmis:document:%d" % doc.id]}
      url = url_for("documents.folder_post",
                    community_id=self.community.slug, folder_id=folder.id)
      response = self.client.post(url, data=data)
      self.assert_200(response)
      assert response.content_type == 'application/zip'

      zipfile = ZipFile(StringIO(response.data))
      assert zipfile.namelist() == [title]

  def test_recursive_zip(self):
    with self.client_login(self.user.email, password='azerty'):
      data = {'action': 'new', 'title': u'my folder'}
      folder = self.community.folder
      url = url_for("documents.folder_post",
                    community_id=self.community.slug, folder_id=folder.id)
      response = self.client.post(url, data=data)
      self.assert_302(response)

      my_folder = folder.children[0]

      title = u"onepage.pdf"
      content_type = "application/pdf"
      data = {
        'file': (self.open_file(title), title, content_type),
        'action': 'upload',
      }
      url = url_for("documents.folder_post",
                    community_id=self.community.slug, folder_id=my_folder.id)
      response = self.client.post(url, data=data)
      self.assert_302(response)

      data = {'action': 'download',
              'object-selected': ["cmis:folder:%d" % my_folder.id]}
      url = url_for("documents.folder_post",
                    community_id=self.community.slug, folder_id=folder.id)
      response = self.client.post(url, data=data)
      self.assert_200(response)
      assert response.content_type == 'application/zip'

      zipfile = ZipFile(StringIO(response.data))
      assert zipfile.namelist() == ['my folder/' + title]

  def test_document_send_by_mail(self):
    mail = self.app.extensions['mail']
    folder = self.community.folder

    attachment = 'Content-Disposition: attachment; filename="{}"'.format
    attachment_utf8 = 'Content-Disposition: attachment;\n filename*="UTF8\'\'{}"'.format

    with self.client_login(self.user.email, password='azerty'):
      # upload files
      for filename in (u'ascii title.txt', u'utf-8 est arrivé!.txt'):
        content_type = "text/plain"
        data = {
          'file': (StringIO('file content'), filename, content_type),
          'action': 'upload',
        }
        url = url_for("documents.folder_post",
                      community_id=self.community.slug, folder_id=folder.id)
        self.client.post(url, data=data)

      ascii_doc = folder.children[0]
      unicode_doc = folder.children[1]

      def get_send_url(doc_id):
        return url_for('documents.document_send',
                       community_id=self.community.slug,
                       doc_id=doc_id)

      # mail ascii filename
      with mail.record_messages() as outbox:
        url = get_send_url(ascii_doc.id)
        rv = self.client.post(url, data={'recipient': 'dest@example.com',
                                         'message': u'Voilà un fichier'})
        self.assertEquals(rv.status_code, 302, "expected 302, got:" + rv.status)
        assert len(outbox) == 1
        msg = outbox[0]
        self.assertEquals(
          msg.subject,
          u'[Abilian Test] Unknown sent you a file'
        )
        assert msg.recipients == [u'dest@example.com']
        expected_disposition = attachment('ascii title.txt')
        msg = str(msg)
        assert expected_disposition in msg

      # mail unicode filename
      with mail.record_messages() as outbox:
        url = get_send_url(unicode_doc.id)
        rv = self.client.post(url, data={'recipient': 'dest@example.com',
                                         'message': u'Voilà un fichier'})
        assert rv.status_code == 302
        assert len(outbox) == 1
        msg = outbox[0]
        self.assertEquals(
          msg.subject,
          u'[Abilian Test] Unknown sent you a file'
        )
        self.assertEquals(msg.recipients, [u'dest@example.com'])
        expected_disposition = attachment_utf8(
          'utf-8%20est%20arriv%C3%A9%21.txt')
        msg = str(msg)
        self.assertTrue(expected_disposition in msg,
                        (expected_disposition, msg))


class TestPathIndexable(unittest.TestCase):
  class MockPath(PathAndSecurityIndexable):
    def __init__(self, id, parent=None):
      self.id = id
      self.parent = parent


  def setUp(self):
    id_gen = count()
    obj = self.MockPath(id_gen.next())
    obj = self.MockPath(id_gen.next(), parent=obj)
    obj = self.MockPath(id_gen.next(), parent=obj)
    self.obj = self.MockPath(id_gen.next(), parent=obj)

  def test_iter_to_root(self):
    assert [o.id for o in self.obj._iter_to_root()] == [3, 2, 1, 0]
    assert [o.id for o in self.obj._iter_to_root(skip_self=True)] == [2, 1, 0]

  def test_indexable_parent_ids(self):
    self.assertEquals(self.obj._indexable_parent_ids,
                      u'/0/1/2')



from flask import url_for
from abilian.core.models.subjects import User
from abilian.sbe.testing import BaseTestCase


class TestSocialViews(BaseTestCase):

  init_data = True
  no_login = True

  # Tests start here
  def test_home(self):
    response = self.client.get(url_for("social.home"))
    self.assert_200(response)
    self.assert_valid(response)

  def test_users(self):
    response = self.client.get(url_for("social.users"))
    self.assert_200(response)
    self.assert_valid(response)

  def test_groups(self):
    response = self.client.get(url_for("social.groups"))
    self.assert_200(response)
    self.assert_valid(response)

  def test_user(self):
    user = User.query.all()[0]
    response = self.client.get(url_for("social.user", user_id=user.id))
    self.assert_200(response)
    self.assert_valid(response)



# coding=utf-8
"""
"""
from __future__ import absolute_import

from datetime import datetime, timedelta

import mock
from pytz import UTC

from abilian.testing import BaseTestCase
from abilian.core.models.subjects import User
from abilian.sbe.apps.documents import lock
from abilian.sbe.apps.documents.lock import Lock

def test_lock():
  date = datetime(2015, 10, 22, 14, 58, 42, tzinfo=UTC)
  l = Lock(user_id=3, user=u'Joe Smith', date=date)

  d = l.as_dict()
  assert d == dict(user_id=3, user=u'Joe Smith',
                   date=u'2015-10-22T14:58:42+00:00')
  l = Lock.from_dict(d)
  assert l.user_id == 3
  assert l.user == u'Joe Smith'
  assert l.date == date


class LockTestCase(BaseTestCase):

  SERVICES = ('security',)

  def test_lock(self):
    session = self.app.db.session()
    user = User(email=u'test@example.com',
                first_name=u'Joe', last_name=u'Smith',
                can_login=True)
    other = User(email=u'other@exemple.com')
    session.add(user)
    session.add(other)
    session.commit()

    # set 30s lifetime
    self.app.config['SBE_LOCK_LIFETIME'] = 30
    dt_patcher = mock.patch.object(lock, 'utcnow',
                                   mock.Mock(wraps=lock.utcnow))
    with dt_patcher as mocked:
      created_at = datetime(2015, 10, 22, 14, 58, 42, tzinfo=UTC)
      mocked.return_value = created_at

      self.login(user)
      l = Lock.new()
      assert l.user_id == user.id
      assert l.user == u'Joe Smith'
      assert l.date == created_at
      assert l.is_owner()
      assert l.is_owner(user)
      assert not l.is_owner(other)

      assert l.lifetime == 30
      mocked.return_value = created_at + timedelta(seconds=40)
      assert l.expired

      mocked.return_value = created_at + timedelta(seconds=20)
      assert not l.expired

      self.login(other)
      assert not l.is_owner()



# coding=utf-8
"""
"""
from __future__ import absolute_import
from datetime import date
from itertools import groupby, islice

import whoosh
import whoosh.query as wq
from sqlalchemy.orm import joinedload
from flask import g, render_template, current_app
from flask_babel import format_date

from abilian.web import url_for
from abilian.web.action import actions
from abilian.sbe.apps.documents.models import Document, icon_for
from abilian.sbe.apps.forum.models import Thread
from abilian.sbe.apps.communities.blueprint import Blueprint

from .util import get_recent_entries
from .presenters import ActivityEntryPresenter

wall = Blueprint("wall", __name__,
                 url_prefix="/wall",
                 template_folder="templates")
route = wall.route


@wall.url_value_preprocessor
def set_current_tab(endpoint, values):
  g.current_tab = 'wall'


@route('/')
def index():
  actions.context['object'] = g.community._model
  entries = get_recent_entries(20, community=g.community)
  entries = ActivityEntryPresenter.wrap_collection(entries)
  return render_template("wall/index.html", entries=entries)


@route('/files')
def files():
  community = g.community._model
  actions.context['object'] = community

  all_files = []
  if community.has_forum:
    all_files += get_attachments_from_forum(community)

  if community.has_documents:
    all_files += get_attachments_from_dms(community)

  all_files = sorted(all_files, key=lambda doc: doc.date)
  all_files = islice(reversed(all_files), 0, 50)
  grouped_docs = group_monthly(all_files)

  return render_template("wall/files.html", grouped_docs=grouped_docs)


class Attachment(object):
  def __init__(self, url, name, owner, date, content_length, content_type):
    self.url = url
    self.name = name
    self.owner = owner
    self.date = date
    self.content_length = content_length
    self.content_type = content_type

  @property
  def icon(self):
    return icon_for(self.content_type)


def get_attachments_from_forum(community):
  all_threads = Thread.query \
    .filter(Thread.community_id == community.id) \
    .options(joinedload('posts')) \
    .options(joinedload('posts.attachments')) \
    .order_by(Thread.created_at.desc()).all()

  posts_with_attachments = []
  for thread in all_threads:
    for post in thread.posts:
      if getattr(post, 'attachments', None):
        posts_with_attachments.append(post)

  posts_with_attachments.sort(key=lambda post: post.created_at)
  posts_with_attachments.reverse()

  attachments = []
  for post in posts_with_attachments:
    for att in post.attachments:
      url = current_app.default_view.url_for(att)
      attachment = Attachment(url, att.name, unicode(att.owner),
                              att.created_at, att.content_length, att.content_type)
      attachments.append(attachment)

  return attachments


# FIXME: significant performance issues here, needs major refactoring
def get_attachments_from_dms(community):
  svc = current_app.services['indexing']
  filters = wq.And([wq.Term('community_id', community.id),
                    wq.Term('object_type', Document.entity_type)])
  sortedby = whoosh.sorting.FieldFacet('created_at', reverse=True)
  documents = svc.search(u'', filter=filters, sortedby=sortedby, limit=50)

  attachments = []
  for doc in documents:
    url = url_for(doc)
    attachment = Attachment(url, doc['name'], doc['owner_name'],
                            doc['created_at'],
                            doc.get('content_length', None),
                            doc.get('content_type', u''),)
    attachments.append(attachment)

  return attachments


# FIXME: copy/pasted then hacked from forum/views.py
def group_monthly(objects):
  # We're using Python's groupby instead of SA's group_by here
  # because it's easier to support both SQLite and Postgres this way.
  def grouper(entity):
    return entity.date.year, entity.date.month

  def format_month(year, month):
    month = format_date(date(year, month, 1), "MMMM").capitalize()
    return u"%s %s" % (month, year)

  grouped = groupby(objects, grouper)
  grouped = [(format_month(year, month), list(objs))
             for (year, month), objs in grouped]
  return grouped



# coding=utf-8
"""
"""
from __future__ import absolute_import

from abilian.i18n import _l
from abilian.web import nav
from flask import g, Blueprint as BaseBlueprint
from werkzeug.exceptions import NotFound

from . import security
from .models import Community
from .presenters import CommunityPresenter


class Blueprint(BaseBlueprint):
  """
  Blueprint for community based views.

  It sets g.community and perform access verification for the traversed
  community.
  """
  _BASE_URL_PREFIX = '/communities'
  _ROUTE_PARAM = '<string:community_id>'

  def __init__(self, *args, **kwargs):
    url_prefix = kwargs.get('url_prefix', '')

    if kwargs.pop('set_community_id_prefix', True):
      if (not url_prefix) or url_prefix[0] != '/':
        url_prefix = '/' + url_prefix
      url_prefix = self._ROUTE_PARAM + url_prefix

    if not url_prefix.startswith(self._BASE_URL_PREFIX):
      if (not url_prefix) or url_prefix[0] != '/':
        url_prefix = '/' + url_prefix
      url_prefix = self._BASE_URL_PREFIX + url_prefix

    if url_prefix[-1] == '/':
      url_prefix = url_prefix[:-1]
    kwargs['url_prefix'] = url_prefix

    BaseBlueprint.__init__(self, *args, **kwargs)
    self.url_value_preprocessor(pull_community)
    self.before_request(check_access)


def check_access():
  if hasattr(g, 'community'):
    # communities.index is not inside a community, for example
    security.check_access(g.community)


def pull_community(endpoint, values):
  """
  url_value_preprocessor function
  """
  g.nav['active'] = 'section:communities'
  g.breadcrumb.append(
    nav.BreadcrumbItem(label=_l(u'Communities'),
                       url=nav.Endpoint('communities.index')))

  try:
    slug = values.pop('community_id')
    community = Community.query.filter(Community.slug == slug).first()
    if community:
      g.community = CommunityPresenter(community)
      wall_url = nav.Endpoint('wall.index', community_id=community.slug)
      breadcrumb_item = nav.BreadcrumbItem(label=community.name,
                                           url=wall_url)
      g.breadcrumb.append(breadcrumb_item)
    else:
      raise NotFound()
  except KeyError:
    pass



# coding=utf-8
"""
Static configuration for the application.

TODO: add more (runtime) flexibility in plugin discovery, selection
and activation.
"""
from __future__ import absolute_import

import os
import sys
import subprocess
import logging

from flask import current_app
from flask_script import Manager, Command
import jinja2
from werkzeug.serving import BaseWSGIServer
from pathlib import Path
from abilian.app import Application as BaseApplication
from abilian.core.extensions import db
from abilian.core.commands import setup_abilian_commands
from abilian.services import converter
from abilian.core.celery import (
  FlaskLoader as CeleryBaseLoader, FlaskCelery as BaseCelery
)

from .apps.documents.repository import repository
from .extension import sbe


# Used for side effects, do not remove


__all__ = ['create_app', 'db']

logger = logging.getLogger(__name__)

def create_app(config=None):
  return Application(config=config)


command_manager = Manager(create_app)
setup_abilian_commands(command_manager)

# loader to be used by celery workers
class CeleryLoader(CeleryBaseLoader):
  flask_app_factory = 'abilian.sbe.app.create_app'


celery = BaseCelery(loader=CeleryLoader)


class Application(BaseApplication):

  APP_PLUGINS = BaseApplication.APP_PLUGINS + (
    "abilian.sbe.apps.main",
    "abilian.sbe.apps.notifications",
    "abilian.sbe.apps.preferences",
    "abilian.sbe.apps.wiki",
    "abilian.sbe.apps.wall",
    "abilian.sbe.apps.documents",
    "abilian.sbe.apps.forum",
    "abilian.sbe.apps.communities",
    "abilian.sbe.apps.social",
    "abilian.sbe.apps.preferences",
  )

  script_manager = command_manager

  def __init__(self, name='abilian_sbe', config=None, **kwargs):
    BaseApplication.__init__(self, name, config=config, **kwargs)
    self.register_jinja_loaders(
      jinja2.PackageLoader('abilian.sbe', 'templates')
    )

  def init_extensions(self):
    BaseApplication.init_extensions(self)
    sbe.init_app(self)
    repository.init_app(self)
    converter.init_app(self)


# SBE demo app bootstrap stuff
_SBE_DEMO_SCRIPT = u'''\
#!{BIN_DIR}/python
from __future__ import absolute_import

import sys
from abilian.sbe.app import command_entry_point

sys.exit(command_entry_point())
'''


_BASE_SERVER_ACTIVATE = BaseWSGIServer.server_activate


def _on_http_server_activate(self, *args, **kwargs):
  """
  This function is used as to monkey patch BaseWSGIServer.server_activate
  during `setup_sbe_demo`.
  """
  _BASE_SERVER_ACTIVATE(self, *args, **kwargs)
  # now we are listening to socket
  host, port = self.server_address
  if host == u'0.0.0.0':
    # chrome is not ok with 0.0.0.0
    host = u'localhost'
  url = 'http://{host}:{port}/setup'.format(host=host, port=port)

  if sys.platform == "win32":
    os.startfile(url)
  else:
    opener ="open" if sys.platform == "darwin" else "xdg-open"
    subprocess.call([opener, url])


# run with python -m abilian.sbe.app setup_sbe_app
def setup_sbe_app():
  """
  Basic set up SBE application. Must be run inside a virtualenv.

  Will create `abilian_sbe` script, run a local server and open browser on app's
  setup wizard.
  """
  logger = logging.getLogger('sbe_demo')
  logger.setLevel(logging.INFO)

  if 'VIRTUAL_ENV' not in os.environ:
    logger.error('Not in a virtualenv! Aborting.')
    return 1

  bin_dir = Path(sys.prefix) / u'bin'

  if not bin_dir.exists() and bin_dir.is_dir():
    logger.error('%s doesn\'t exists or is not a directory. Aborting',
                 repr(unicode(bin_dir)))
    return 1

  script_file = bin_dir / u'abilian_sbe'
  if script_file.exists():
    logger.info('%s already exists. Skipping creation.',
                repr(unicode(script_file)))
  else:
    with script_file.open('w') as out:
      logger.info('Create script: "%s".', repr(unicode(script_file)))
      content = _SBE_DEMO_SCRIPT.format(BIN_DIR=unicode(bin_dir))
      out.write(content)
    script_file.chmod(0755) # 0755: -rwxr-xr-x

  current_app.config['PRODUCTION'] = True
  current_app.config['DEBUG'] = False
  current_app.config['ASSETS_DEBUG'] = False
  current_app.config['SITE_NAME'] = u'Abilian SBE'
  current_app.config['MAIL_SENDER'] = u'abilian-sbe-app@example.com'

  logger.info('Prepare CSS & JS files')
  command_manager.handle('abilian_sbe', ['assets', 'build'])
  # disabled init config: only if not running setupwizard
  # command_manager.handle('abilian_sbe', ['config', 'init'])

  # patch server used to launch browser here immediately after socket opened
  BaseWSGIServer.server_activate = _on_http_server_activate
  return command_manager.handle('abilian_sbe', ['run', '--hide-config'])


def command_entry_point():
  command_manager.run(
      commands={'setup_sbe_app': Command(setup_sbe_app)},
  )


if __name__ == '__main__':
  command_entry_point()



# coding=utf-8
"""
First cut at a notification system.
"""
from __future__ import absolute_import

from werkzeug.exceptions import MethodNotAllowed
from flask import current_app as app, request
from flask.ext.login import current_user
from abilian.i18n import render_template_i18n
from abilian.core.extensions import db, csrf
from abilian.core.models.subjects import User
from abilian.services.auth.views import get_token_status
from abilian.sbe.apps.communities.security import require_admin

from abilian.sbe.apps.notifications import TOKEN_SERIALIZER_NAME
from ..tasks.social import send_daily_social_digest_to, make_message
from . import notifications

__all__ = []

route = notifications.route

@require_admin
@route("/debug/social/")
def debug_social():
  """Send a digest to current user, or user with given email.

  Also displays the email in the browser as a result.
  """
  email = request.args.get('email')
  if email:
    user = User.query.filter(User.email == email).one()
  else:
    email = current_user.email
    user = current_user

  msg = make_message(user)

  status = send_daily_social_digest_to(user)

  if status:
    return msg.html
  else:
    return "No message sent."


@route("/unsubscribe_sbe/<token>/", methods=['GET', 'POST'])
@csrf.exempt
def unsubscribe_sbe(token):
  expired, invalid, user = get_token_status(token, TOKEN_SERIALIZER_NAME)
  if expired or invalid:
    return render_template_i18n("notifications/invalid-token.html")

  if request.method == 'GET':
    return render_template_i18n("notifications/confirm-unsubscribe.html",
                           token=token)

  elif request.method == 'POST':
    preferences = app.services['preferences']
    preferences.set_preferences(user, **{'sbe:notifications:daily': False})
    db.session.commit()
    return render_template_i18n("notifications/unsubscribed.html",
                           token=token)
  else:
    raise MethodNotAllowed()



# coding=utf-8
"""
Celery tasks related to document transformation and preview.
"""
from __future__ import absolute_import

import mailbox
from os.path import expanduser
import re
from pathlib import Path

import bleach
import chardet
from itsdangerous import URLSafeSerializer

from flask import current_app, g
from flask_mail import Message
from flask_babel import get_locale
from celery import shared_task
from celery.utils.log import get_task_logger
from celery.schedules import crontab

from abilian.web import url_for
from abilian.core.celery import periodic_task
from abilian.core.signals import activity
from abilian.core.extensions import mail, db
from abilian.core.models.subjects import User
from abilian.i18n import _l, render_template_i18n

from .forms import ALLOWED_ATTRIBUTES, ALLOWED_TAGS, ALLOWED_STYLES
from .models import Thread, PostAttachment

MAIL_REPLY_MARKER = _l(u'_____Write above this line to post_____')

# logger = logging.getLogger(__package__)
# Celery logger
logger = get_task_logger(__name__)


def init_app(app):
  global check_maildir
  if app.config['INCOMING_MAIL_USE_MAILDIR']:
    make_task = periodic_task(run_every=crontab(minute='*', ),)
    check_maildir = make_task(check_maildir)


@shared_task()
def send_post_by_email(post_id):
  """
  Send a post to community members by email.
  """
  from .models import Post

  with current_app.test_request_context('/send_post_by_email'):
    post = Post.query.get(post_id)
    if post is None:
      # deleted after task queued, but before task run
      return

    thread = post.thread
    community = thread.community
    logger.info("Sending new post by email to members of community %r",
                community.name)

    CHUNK_SIZE = 20
    members_id = [member.id for member in community.members
                  if member.can_login]
    chunk = []
    for idx, member_id in enumerate(members_id):
      chunk.append(member_id)
      if idx % CHUNK_SIZE == 0:
        batch_send_post_to_users.apply_async((post.id, chunk))
        chunk = []

    if chunk:
      batch_send_post_to_users.apply_async((post.id, chunk))


@shared_task(max_retries=10, rate_limit='12/m')
def batch_send_post_to_users(post_id, members_id, failed_ids=None):
  """
  task run from send_post_by_email; auto-retry for mails that could not be
  successfully sent.

  Task default rate limit is 6/min.: there is at least 5 seconds between 2
  batches.

  During retry, if all `members_id` fails again, the task is retried
  5min. later, then 10, 20, 40... up to 10 times before giving up. This ensures
  retries up to approximatively 3 days and 13 hours after initial attempt
  (geometric series is: 5min * (1-2**10) / 1-2) = 5115 mins).
  """
  from .models import Post

  if not members_id:
    return

  post = Post.query.get(post_id)
  if post is None:
    # deleted after task queued, but before task run
    return

  failed = set()
  successfully_sent = []
  thread = post.thread
  community = thread.community
  user_filter = (User.id.in_(members_id) if len(members_id) > 1
                 else User.id == members_id[0])
  users = User.query.filter(user_filter).all()

  for user in users:
    try:
      with current_app.test_request_context('/send_post_by_email'):
        send_post_to_user(community, post, user)
    except:
      failed.add(user.id)
    else:
      successfully_sent.append(user.id)

  if failed:
    if failed_ids is not None:
      failed_ids = set(failed_ids)

    if failed == failed_ids:
      # 5 minutes * (2** retry count)
      countdown = 300 * 2 ** batch_send_post_to_users.request.retries
      batch_send_post_to_users.retry([post_id, list(failed)],
                                     countdown=countdown)
    else:
      batch_send_post_to_users.apply_async([post_id, list(failed)])

  return {'post_id': post_id,
          'successfully_sent': successfully_sent,
          'failed': list(failed),}


def build_reply_email_address(name, post, member, domain):
  """
    Builds a reply-to email address
    embedding the locale, thread_id and user.id

  :param name: (str)    first part of an email address
  :param post: Post()   to get post.thread_id
  :param member: User() to get user.id
  :param domain: (str)  the last domain name of the email address
  :return: (unicode)    reply address for forum in the form
    test+IjEvMy8yLzQi.xjE04-4S0IzsdicTHKTAqcqa1fE@testcase.app.tld

    the hidden parameters are:  locale/post.thread_id/user.id
  """
  tag = current_app.config['MAIL_ADDRESS_TAG_CHAR']
  key = current_app.config['SECRET_KEY']
  locale = get_locale()
  serializer = URLSafeSerializer(key)
  uid = u'/'.join([str(locale), str(post.thread_id), str(member.id)])
  return name + tag + serializer.dumps(uid.decode('utf-8')) + u'@' + domain


def extract_email_destination(address):
  """
    Returns the values encoded in the email address.
  :param address: similar to test+IjEvMy8yLzQi.xjE04-4S0IzsdicTHKTAqcqa1fE@testcase.app.tld
  :return: List() of splitted values
  """
  name = address.rsplit('@', 1)[0]
  tag = current_app.config['MAIL_ADDRESS_TAG_CHAR']
  key = current_app.config['SECRET_KEY']
  serializer = URLSafeSerializer(key)
  values = name.rsplit(tag, 1)[1]
  return serializer.loads(values).split('/')


def has_subtag(address):
  """
    Returns True if a subtag (defined in the config.py as 'MAIL_ADDRESS_TAG_CHAR')
    was found in the name part of the address
    :param address: email adress
    :return: Boolean
  """
  name = address.rsplit('@', 1)[0]
  tag = current_app.config['MAIL_ADDRESS_TAG_CHAR']
  return (tag in name)


def send_post_to_user(community, post, member):
  recipient = member.email
  subject = u'[%s] %s' % (community.name, post.title)
  config = current_app.config
  sender = config.get('BULK_MAIL_SENDER', config['MAIL_SENDER'])
  SBE_FORUM_REPLY_BY_MAIL = config.get('SBE_FORUM_REPLY_BY_MAIL', False)
  SERVER_NAME = config.get('SERVER_NAME', u'example.com')
  list_id = u'"{} forum" <forum.{}.{}>'.format(community.name,
                                               community.slug, SERVER_NAME)
  forum_url = url_for('forum.index', community_id=community.slug,
                      _external=True)
  forum_archive = url_for('forum.archives', community_id=community.slug,
                          _external=True)

  extra_headers = {
    'List-Id': list_id,
    'List-Archive': u'<{}>'.format(forum_archive),
    'List-Post': '<{}>'.format(forum_url),
    'X-Auto-Response-Suppress': 'All',
    'Auto-Submitted': 'auto-generated',
  }

  if SBE_FORUM_REPLY_BY_MAIL and config['MAIL_ADDRESS_TAG_CHAR'] is not None:
    name = sender.rsplit('@', 1)[0]
    domain = sender.rsplit('@', 1)[1]
    replyto = build_reply_email_address(name, post, member, domain)
    msg = Message(subject, sender=sender, recipients=[recipient],
                  reply_to=replyto, extra_headers=extra_headers)
  else:
    msg = Message(subject, sender=sender, recipients=[recipient],
                  extra_headers=extra_headers)

  msg.body = render_template_i18n(
    "forum/mail/new_message.txt",
    community=community, post=post, member=member,
    MAIL_REPLY_MARKER=MAIL_REPLY_MARKER,
    SBE_FORUM_REPLY_BY_MAIL=SBE_FORUM_REPLY_BY_MAIL,
  )
  msg.html = render_template_i18n(
    "forum/mail/new_message.html",
    community=community, post=post, member=member,
    MAIL_REPLY_MARKER=MAIL_REPLY_MARKER,
    SBE_FORUM_REPLY_BY_MAIL=SBE_FORUM_REPLY_BY_MAIL,
  )
  logger.debug("Sending new post by email to %r", member.email)

  try:
    mail.send(msg)
  except:
    logger.error("Send mail to user failed",
                 exc_info=True)  # log to sentry if enabled


def extract_content(payload, marker):
  """Search the payload for marker, return content up to marker."""
  index = payload.rfind(marker)
  content = payload[:index]
  return content


def validate_html(payload):
  return bleach.clean(payload, tags=ALLOWED_TAGS,
                      attributes=ALLOWED_ATTRIBUTES,
                      styles=ALLOWED_STYLES, strip=True).strip()


def add_paragraph(newpost):
  """
    Adds surrounding <p>newpost</p> if necessary
  """
  newpost = newpost.strip()
  if not newpost.startswith(u'<p>'):
    newpost = u'<p>' + newpost + u'</p>'
  return newpost


def clean_html(newpost):
  clean = re.sub(r"(<blockquote.*?<p>.*?</p>.*?</blockquote>)",
                 '', newpost, flags=re.MULTILINE | re.DOTALL)
  clean = re.sub(r"(<br>.*?<a href=.*?/a>.*?:<br>)",
                 '', clean, flags=re.MULTILINE | re.DOTALL)
  return clean


def decode_payload(part):
  # Get the payload and decode (base64 & quoted printable)
  payload = part.get_payload(decode=True)
  if not (isinstance(payload, unicode)):
    # What about other encodings? -> using chardet
    if part.get_content_charset() is None:
      found = chardet.detect(payload)
      payload = payload.decode(found['encoding'])
    else:
      payload = payload.decode(part.get_content_charset())
  return payload


def process(message, marker):
  """
    Check the message for marker presence and return the text up to it if present

    :raises LookupError otherwise.
    :param message: email.Message()
    :param marker: unicode
    :return: sanitized html upto marker from message
  """
  content = {'plain': u'', 'html': u''}
  attachments = []
  # Iterate all message's parts for text/*
  for part in message.walk():
    content_type = part.get_content_type()
    content_disp = part.get('Content-Disposition')

    if content_type in ['text/plain', 'text/html'] and content_disp is None:
      payload = content[part.get_content_subtype()] + decode_payload(part)
      content[part.get_content_subtype()] = payload

    if content_disp is not None:
      attachments.append(
        {'filename': part.get_filename(),
         'content_type': part.get_content_type(),
         'data': part.get_payload(decode=True)})

  if 'html' in content and marker in content['html']:
    newpost = extract_content(content['html'], marker[:9])
    newpost = add_paragraph(validate_html(newpost))
    newpost = clean_html(newpost)
  elif 'plain' in content and marker in content['plain']:
    newpost = extract_content(content['plain'], marker[:9])
    newpost = add_paragraph(newpost)
  else:
    logger.error('No marker:{} in email'.format(marker))
    raise LookupError('No marker:{} in email'.format(marker))

  return newpost, attachments


@shared_task()
def process_email(message):
  """
  Email.Message object from command line script Run message (parsed email).

  Processing chain extract community thread post member from reply_to persist
  post in db.
  """
  app = current_app._get_current_object()
  # Extract post destination from To: field, (community/forum/thread/member)
  to_address = message['To']

  if not (has_subtag(to_address)):
    logger.info('Email %r has no subtag, skipping...', to_address)
    return False

  try:
    infos = extract_email_destination(to_address)
    locale = infos[0]
    thread_id = infos[1]
    user_id = infos[2]
  except:
    logger.error('Recipient %r cannot be converted to locale/thread_id/user.id',
                 to_address,
                 exc_info=True)
    return False

  # Translate marker with locale from email address
  rq_headers = [('Accept-Language', locale)]
  with app.test_request_context('/process_email', headers=rq_headers):
    marker = unicode(MAIL_REPLY_MARKER)

  # Extract text and attachments from message
  try:
    newpost, attachments = process(message, marker)
  except:
    logger.error('Could not Process message', exc_info=True)
    return False

  # Persist post
  with current_app.test_request_context('/process_email', headers=rq_headers):
    g.user = User.query.get(user_id)
    thread = Thread.query.get(thread_id)
    community = thread.community
    # FIXME: check membership, send back an informative email in case of an error
    post = thread.create_post(body_html=newpost)
    obj_meta = post.meta.setdefault('abilian.sbe.forum', {})
    obj_meta['origin'] = u'email'
    obj_meta['send_by_email'] = True
    activity.send(app, actor=g.user, verb='post', object=post, target=community)

    if len(attachments) > 0:
      for desc in attachments:
        attachment = PostAttachment(name=desc['filename'])
        attachment.post = post
        attachment.set_content(desc['data'], desc['content_type'])
        db.session.add(attachment)
    db.session.commit()

  # Notify all parties involved
  send_post_by_email.delay(post.id)
  return True


def check_maildir():
  """
  Check the MailDir for emails to be injected in Threads.

  This task is registered only if `INCOMING_MAIL_USE_MAILDIR` is True. By
  default it is run every minute.
  """
  home = expanduser('~')
  maildirpath = str(Path(home) / 'Maildir')
  src_mdir = mailbox.Maildir(maildirpath,
                             factory=mailbox.MaildirMessage)

  src_mdir.lock()  # Useless but recommended if old mbox is used by error

  try:
    for key, message in src_mdir.iteritems():
      processed = process_email(message)

      # delete the message if all went fine
      if processed:
        del src_mdir[key]

  finally:
    src_mdir.close()  # Flushes all changes to disk then unlocks



# coding=utf-8
"""
"""
from __future__ import absolute_import

import logging

from flask import g
from flask_login import current_user
import whoosh.query as wq
import whoosh.fields as wf

from .models import Membership


logger = logging.getLogger(__name__)

_COMMUNITY_CONTENT_FIELDNAME = 'is_community_content'
_COMMUNITY_CONTENT_FIELD = wf.BOOLEAN()

_COMMUNITY_ID_FIELD = wf.NUMERIC(numtype=int, bits=64, signed=False,
                                        stored=True, unique=False)
_COMMUNITY_SLUG_FIELD = wf.ID(stored=True)

_FIELDS = [
  (_COMMUNITY_CONTENT_FIELDNAME, _COMMUNITY_CONTENT_FIELD),
  ('community_id', _COMMUNITY_ID_FIELD),
  ('community_slug', _COMMUNITY_SLUG_FIELD),
]

def init_app(app):
  """
  Add community fields to indexing service schema
  """
  indexing = app.services['indexing']
  indexing.register_search_filter(filter_user_communities)
  indexing.register_value_provider(mark_non_community_content)

  for name, schema in indexing.schemas.items():
    for fieldname, field in _FIELDS:
      if fieldname in schema:
        if schema[fieldname] is not field:
          logger.warning('Field "%s" already in schema %r, replacing with '
                         'expected fieldtype instance', fieldname, schema)
          del schema._fields[fieldname]
        else:
          continue

      schema.add(fieldname, field)


def filter_user_communities():
  if g.is_manager:
    return None

  filter_q = wq.Term(_COMMUNITY_CONTENT_FIELDNAME, False)

  if not current_user.is_anonymous():
    ids = Membership.query\
        .filter(Membership.user == current_user)\
        .order_by(Membership.community_id.asc())\
        .values(Membership.community_id)
    communities = [wq.Term('community_id', i[0]) for i in ids]

    if communities:
      communities = wq.And(
        [wq.Term(_COMMUNITY_CONTENT_FIELDNAME, True),
         wq.Or(communities)])
      filter_q = wq.Or([filter_q, communities])

  return filter_q


def mark_non_community_content(document, obj):
  if _COMMUNITY_CONTENT_FIELDNAME not in document:
    document[_COMMUNITY_CONTENT_FIELDNAME] = \
        getattr(obj, _COMMUNITY_CONTENT_FIELDNAME, False)

  return document



# coding=utf-8
"""
"""
from __future__ import absolute_import

import logging
import pkgutil
from cgi import escape

from flask import (
    request, redirect, g, jsonify,
    current_app, Response, render_template, flash
)
import sqlalchemy as sa
from sqlalchemy.sql.expression import or_, and_, func, asc, desc, nullslast
from abilian.i18n import _, _l
from abilian.core.models.subjects import User
from abilian.core.extensions import db
from abilian.web import url_for
from abilian.web.views import default_view, ObjectEdit
from abilian.web.views.images import user_photo_url
from abilian.web.filters import age
from werkzeug.exceptions import InternalServerError

from abilian.sbe.apps.communities.models import Membership
from abilian.sbe.apps.wall.views import get_recent_entries
from abilian.sbe.apps.wall.presenters import ActivityEntryPresenter
from ..forms import UserProfileForm, UserProfileViewForm
from .social import social
from .util import Env

logger = logging.getLogger(__name__)

DEFAULT_USER_MUGSHOT = pkgutil.get_data(
  'abilian.sbe',
  'static/images/silhouette_unknown.png')


def make_tabs(user):
  return [
    dict(id='profile', label=_(u'Profile'), link=url_for(user, tab='profile')),
    #dict(id='conversations', label=_(u'Conversations'), link=url_for(user), is_online=True),
    dict(id='documents', label=_(u'Documents'), link=url_for(user, tab='documents')),
    dict(id='images', label=_(u'Images'), link=url_for(user, tab='images')),
    dict(id='audit', label=_(u'Audit'), link=url_for(user, tab='audit')),
  ]


@social.route("/users/")
def users():
  query = request.args.get("query")

  if query:
    query = query.replace("%", " ")
    q = or_(User.first_name.like("%" + query + "%"),
            User.last_name.like("%" + query + "%"))
    users = User.query.filter(q).limit(100).all()
  else:
    users = User.query.limit(100).all()
  ctx = dict(users=users)
  return render_template("social/users.html", **ctx)


@social.route("/users/dt_json")
def users_dt_json():
  """JSON call to fill a DataTable. Needs some refactoring."""
  args = request.args

  length = int(args.get("iDisplayLength", 0))
  start = int(args.get("iDisplayStart", 0))
  sort_col = int(args.get("iSortCol_0", 1))
  sort_dir = args.get("sSortDir_0", "asc")
  echo = int(args.get("sEcho", 0))
  search = args.get("sSearch", "").replace("%", "").lower()

  end = start + length

  q = User.query
  total_count = q.count()

  if search:
    # TODO: gérer les accents
    filter = or_(func.lower(User.first_name).like("%" + search + "%"),
                 func.lower(User.last_name).like("%" + search + "%"),)
    q = q.filter(filter)\
         .reset_joinpoint()

  count = q.count()
  SORT_COLS = {
    1: [],  # will be set to [User.last_name, User.first_name]
    2: [User.created_at],
    3: [User.last_active],
  }
  columns = list(SORT_COLS.get(sort_col, []))
  columns.extend([func.lower(User.last_name), func.lower(User.first_name)])

  direction = asc if sort_dir == 'asc' else desc
  order_by = map(direction, columns)

  # sqlite does not support 'NULLS FIRST|LAST' in ORDER BY clauses
  engine = q.session.get_bind(User.__mapper__)
  if engine.name != 'sqlite':
    order_by[0] = nullslast(order_by[0])

  q = q.order_by(*order_by)

  users = q.slice(start, end).all()

  data = []
  MUGSHOT_SIZE = 45
  for user in users:
    # TODO: this should be done on the browser.
    user_url = url_for(".user", user_id=user.id)
    mugshot = user_photo_url(user, size=MUGSHOT_SIZE)
    name = escape(getattr(user, "name") or "")

    cell0 = (u'<a href="{url}"><img src="{src}" width="{size}" height="{size}">'
             '</a>'.format(url=user_url, src=mugshot, size=MUGSHOT_SIZE))
    cell1 = (u'<div class="info"><a href="{user_url}">{name}</a> '
             u'</div>'.format(**locals()))
    cell2 = age(user.created_at)
    cell3 = age(user.last_active)

    cell4 = '' # TODO: follow / unfollow?
    data.append([cell0, cell1, cell2, cell3])

  result = {
    "sEcho": echo,
    "iTotalRecords": total_count,
    "iTotalDisplayRecords": count,
    "aaData": data,
  }
  return jsonify(result)


@social.route("/users/<int:user_id>")
@default_view(social, User, 'user_id')
def user(user_id):
  security = current_app.services['security']
  user = User.query.get(user_id)

  # FIXME: use user profiles
  view_form = UserProfileViewForm(obj=user)

  communities = [m.community for m in user.communautes_membership]
  if g.user != user and (not security.has_role(g.user, 'manager')):
    # filter visible communautes (ticket 165)
    communities = [c for c in communities if c.has_member(g.user)]

  view_form.communautes._set_data(communities)

  # FIXME
  contact = user
  env = Env(user=user,
            contact=contact,
            view_form=view_form,
            can_edit=can_edit(user),
            tabs=make_tabs(user))

  entries = get_recent_entries(user=user)
  entries = ActivityEntryPresenter.wrap_collection(entries)
  env.activity_entries = entries

  return render_template("social/user.html", **env)


def can_edit(user):
  security = current_app.services.get('security')
  if not security:
    return True

  # TODO: introduce a "self" role?
  return (user == g.user) or security.has_role(g.user, 'admin')


class UserProfileEdit(ObjectEdit):

  Model = User
  Form = UserProfileForm
  pk = 'user_id'
  _message_success = _l(u'Profile edited')

  def init_object(self, args, kwargs):
    args, kwargs = super(UserProfileEdit, self).init_object(args, kwargs)
    self.user = self.obj
    return args, kwargs

  def view_url(self):
    return url_for('.user', user_id=self.user.id)

  def edit(self):
    if not can_edit(self.user):
      return Response(status=403)
    return super(UserProfileEdit, self).edit()

  def handle_commit_exception(self, exc):
    db.session.rollback()
    if isinstance(exc, sa.exc.IntegrityError):
      log_msg = 'Error saving user profile'
    else:
      log_msg = 'Unexpected error while saving user profile'
    logger.error(log_msg, exc_info=True, extra={'stack': True})
    flash(_(u'Error occured'), "error")
    return self.redirect_to_view()


social.route("/users/<int:user_id>/edit")(UserProfileEdit.as_view('user_edit'))


@social.route("/users/<int:user_id>", methods=['POST'])
def user_post(user_id):
  user = User.query.get(user_id)
  action = request.form.get('action')
  return_url = request.form.get('return_url')

  if action == 'follow':
    g.user.follow(user)
  elif action == 'unfollow':
    g.user.unfollow(user)
  else:
    raise Exception("Should not happen")
  db.session.commit()

  if return_url:
    # TODO: security check
    return redirect(return_url)
  else:
    return redirect(url_for(".user_view", user_id=user_id))


#
# Ad-hoc JSON endpoints, used by select boxes
#
@social.route("/users/json")
def users_json():
  q = request.args.get("q").replace("%", " ").lower()

  if not q or len(q) < 2:
    raise InternalServerError()

  query = User.query\
    .filter(or_(func.lower(User.first_name).like(q + "%"),
                func.lower(User.last_name).like(q + "%")))\
    .order_by(func.lower(User.last_name))

  with_membership = request.args.get('with_membership', None)
  if with_membership is not None:
    # provide membership info for a community
    with_membership = int(with_membership)
    query = query\
        .outerjoin(Membership,
                   and_(Membership.user.expression,
                        Membership.community_id == with_membership))\
        .add_columns(Membership.role)

  exclude_community = request.args.get('exclude_community', None)
  if exclude_community is not None:
    exclude_community = int(exclude_community)
    exclude = ~Membership.query\
        .filter(
            Membership.user.expression,
            Membership.community_id == exclude_community)\
        .options(sa.orm.noload('user'),
                 sa.orm.noload('community'))\
        .exists()
    query = query.filter(exclude)

  results = []
  for user in query.all():
    role = None
    if with_membership is not None:
      user, role = user

    if role is not None:
      role = unicode(role)

    item = {'id': user.id,
            'text': u'{} ({})'.format(user.name, user.email),
            'name': user.name,
            'email': user.email,
            'role': role,}
    results.append(item)

  return jsonify({'results': results})



# coding=utf-8
"""

"""
from __future__ import absolute_import

import logging
import fileinput
from email.parser import FeedParser

from flask_script import Manager

from .tasks import process_email, check_maildir

logger = logging.getLogger(__name__)


manager = Manager(description='SBE forum commands',
                  help='SBE forum commands')


@manager.command
def inject_email():
  """
  Reads one email from stdin,
  parse it,
  forward it in a celery task to be persisted.
  """
  parser = FeedParser()
  message = None
  try:
    # iterate over stdin
    for line in fileinput.input('-'):
      parser.feed(line)
  except:
    logger.error('Error during email parsing', exc_info=True)
  finally:
    # close the parser to generate a email.message
    message = parser.close()
    fileinput.close()

  if message:
    # make sure no email.errors are present
    if len(message.defects) == 0:
      process_email.delay(message)
    else:
      logger.error('email has defects')
  else:
    logger.error('no email was parsed from stdin')


@manager.command
def check_email():
  """
  Reads one email from current user Maildir,
  parse it,
  forward it in a celery task to be persisted.
  """

  check_maildir.delay()



# coding=utf-8
"""
"""
from __future__ import absolute_import

import hashlib
import logging
from operator import attrgetter
from functools import wraps
from datetime import datetime
from pathlib import Path
import cStringIO as StringIO
from time import strftime, gmtime

import pytz
import sqlalchemy as sa
from whoosh.searching import Hit
import openpyxl
from openpyxl.writer.write_only import WriteOnlyCell
from werkzeug.exceptions import NotFound, BadRequest, InternalServerError
from flask import (
    render_template, g, redirect, url_for, request, current_app, session,
    flash, jsonify)
from flask_login import current_user, login_required


from abilian.core.extensions import db
from abilian.core.signals import activity
from abilian.core.models.subjects import User
from abilian.core.util import utc_dt
from abilian.services.activity import ActivityEntry
from abilian.services.security import Role
from abilian.web import csrf, nav, views
from abilian.web.views import images as image_views
from abilian.i18n import _, _l

from abilian.sbe.apps.documents.models import Document
from .actions import register_actions
from .forms import CommunityForm
from .models import Community, Membership
from .security import require_admin, require_manage
from .blueprint import Blueprint

__all__ = ['communities']

logger = logging.getLogger(__name__)

EPOCH = datetime.fromtimestamp(0.0, tz=pytz.utc)

def seconds_since_epoch(dt):
  if not dt:
    return 0
  return int((utc_dt(dt) - EPOCH).total_seconds())

communities = Blueprint("communities", __name__,
                        set_community_id_prefix=False,
                        template_folder='templates')
route = communities.route
add_url = communities.add_url_rule
communities.record_once(register_actions)


@communities.record_once
def register_context_processors(state):
  @state.app.context_processor
  def communities_context_processor():
    # helper to get an url for community image
    return dict(community_image_url=image_url)


def tab(tab_name):
  """
  Decorator for view functions to set the current "section" this view
  belongs to.
  """
  def decorator(f):
    @wraps(f)
    def set_current_tab(*args, **kwargs):
      g.current_tab = tab_name
      return f(*args, **kwargs)

    return set_current_tab
  return decorator


def default_view_kw(kw, obj, obj_type, obj_id, **kwargs):
  """
  Helper for using :func:`abilian.web.views.default_view` on objects that
  belongs to a community. This function should be used as `kw_func`::

      @default_view(blueprint, Model, kw_func=default_view_kw)
      @blueprint.route("/<object_id>")
      def view():
         ...

  """
  is_community = obj_type == Community.entity_type
  community_id = kw.get('community_id')

  if is_community or community_id is None:
    # when it's a community, default_view sets community_id to 'id', we want to
    # override with the slug value.
    if obj:
      if isinstance(obj, (Hit, dict)):
        community_id = obj.get('slug' if is_community else 'community_slug')
      elif is_community:
        community_id = obj.slug
      elif community_id is None and hasattr(obj, 'community'):
        try:
          community_id = obj.community.slug
        except AttributeError:
          pass

  if community_id is not None:
    kw['community_id'] = community_id
  else:
    raise ValueError('Cannot find community_id value')

  return kw


#
# Routes
#
@route("/")
@login_required
def index():
  query = Community.query
  sort_order = request.args.get('sort', u'').strip()
  if not sort_order:
    sort_order = session.get('sort_communities_order', 'alpha')

  if sort_order == 'activity':
    query = query.order_by(Community.last_active_at.desc())
  else:
    query = query.order_by(Community.name)

  session['sort_communities_order'] = sort_order

  if not current_user.has_role('admin'):
    # Filter with permissions
    query = query.join(Membership).filter(Membership.user == current_user)

  ctx = dict(my_communities=query.all(), sort_order=sort_order)
  return render_template("community/home.html", **ctx)


@route("/<string:community_id>/")
@views.default_view(communities, Community, 'community_id',
                    kw_func=default_view_kw)
def community():
  return redirect(url_for("wall.index", community_id=g.community.slug))


@route("/json2")
def list_json2():
  """
JSON endpoint, used for filling select boxes dynamically.

  """
  args = request.args
  cls = Community

  q = args.get("q").replace("%", " ")
  if not q or len(q) < 2:
    raise BadRequest()

  query = db.session.query(cls.id, cls.name)
  query = query.filter(cls.name.ilike("%" + q + "%"))\
               .distinct()\
               .order_by(cls.name)\
               .limit(50)
  all = query.all()

  result = {'results': [ { 'id': r[0], 'text': r[1]} for r in all ] }
  return jsonify(result)


# edit views
class BaseCommunityView(object):
  Model = Community
  pk = 'community_id'
  Form = CommunityForm
  base_template = 'community/_base.html'
  decorators = [require_admin]

  def init_object(self, args, kwargs):
    self.obj = g.community._model
    return args, kwargs

  def view_url(self):
    return url_for(self.view_endpoint, community_id=self.obj.slug)

  def get_form_kwargs(self):
    image = self.obj.image
    kwargs = dict()
    if image and 'community' in g:
      setattr(image, 'url', image_url(self.obj, s=500))
      kwargs['image'] = image

    return kwargs


class CommunityEdit(BaseCommunityView, views.ObjectEdit):
  template = 'community/edit.html'
  title = _l("Edit community")
  decorators = views.ObjectEdit.decorators + (require_admin, tab('settings'),)

  def breadcrumb(self):
    return nav.BreadcrumbItem(label=_(u'Settings'),
                              icon='cog',
                              url=nav.Endpoint('communities.settings',
                                               community_id=g.community.slug))

  def get_form_kwargs(self):
    kwargs = views.ObjectEdit.get_form_kwargs(self)
    kwargs.update(BaseCommunityView.get_form_kwargs(self))
    return kwargs

  def before_populate_obj(self):
    form = self.form
    name = form.name.data
    if name != self.obj.name:
      self.obj.rename(name)

    del form.name

    type = form.type.data
    if type != self.obj.type:
      self.obj.type = type
      self.obj.update_roles_on_folder()
    del form.type


add_url("/<string:community_id>/settings",
        view_func=CommunityEdit.as_view(
          'settings',
          view_endpoint='.community',
          message_success=_l(u"Community settings saved successfully.")))


class CommunityCreate(views.ObjectCreate, CommunityEdit):
  title = _l("Create community")
  decorators = views.ObjectCreate.decorators + (require_admin,)
  template = views.ObjectCreate.template
  base_template = views.ObjectCreate.base_template

  def breadcrumb(self):
    return nav.BreadcrumbItem(label=_(u'Create new community'))

  def message_success(self):
    return _(u"Community %(name)s created successfully", name=self.obj.name)


add_url('/new',
        view_func=CommunityCreate.as_view('new', view_endpoint='.community'))


class CommunityDelete(BaseCommunityView, views.ObjectDelete):
  get_form_kwargs = views.ObjectDelete.get_form_kwargs


add_url("/<string:community_id>/destroy",
        methods=['POST'],
        view_func=CommunityDelete.as_view(
          'delete',
          message_success=_l(u"Community destroyed.")))

# Community Image
_DEFAULT_IMAGE = Path(__file__).parent / u'data' / u'community.png'
_DEFAULT_IMAGE_MD5 = hashlib.md5(_DEFAULT_IMAGE.open('rb').read()).hexdigest()
route('/_default_image')(
  image_views.StaticImageView.as_view('community_default_image',
                                      set_expire=True,
                                      image=_DEFAULT_IMAGE,)
)


class CommunityImageView(image_views.BlobView):
  id_arg = 'blob_id'

  def prepare_args(self, args, kwargs):
    community = g.community
    if not community:
      raise NotFound()

    kwargs[self.id_arg] = community.image.id
    # image = open(join(dirname(__file__), "data", "community.png"), 'rb')
    return super(CommunityImageView, self).prepare_args(args, kwargs)


image = CommunityImageView.as_view('image', max_size=500, set_expire=True)
route("/<string:community_id>/image")(image)


def image_url(community, **kwargs):
  """
  return proper URL for image url
  """
  if not community or not community.image:
    kwargs['md5'] = _DEFAULT_IMAGE_MD5
    return url_for('communities.community_default_image', **kwargs)

  kwargs['community_id'] = community.slug
  kwargs['md5'] = community.image.md5
  return url_for('communities.image', **kwargs)


def _members_query():
  """
  Helper used in members views
  """
  last_activity_date = sa.sql.functions.max(ActivityEntry.happened_at)\
                                       .label('last_activity_date')
  memberships = User\
    .query\
    .options(sa.orm.undefer('photo'))\
    .join(Membership)\
    .outerjoin(ActivityEntry,
               sa.sql.and_(ActivityEntry.actor_id == User.id,
                           ActivityEntry.target_id == Membership.community_id))\
    .filter(Membership.community == g.community)\
    .add_columns(Membership.id,
                 Membership.role,
                 last_activity_date,)\
    .group_by(User, Membership.id, Membership.role)\
    .order_by(User.last_name.asc(), User.first_name.asc())

  return memberships


@route("/<string:community_id>/members")
@tab('members')
def members():
  g.breadcrumb.append(nav.BreadcrumbItem(
    label=_(u'Members'),
    url=nav.Endpoint('communities.members', community_id=g.community.slug))
  )
  memberships = _members_query().all()

  return render_template("community/members.html",
                         seconds_since_epoch=seconds_since_epoch,
                         memberships=memberships,
                         csrf_token=csrf.field())


@route("/<string:community_id>/members", methods=["POST"])
@csrf.protect
@require_manage
def members_post():
  community = g.community._model
  action = request.form.get("action")

  user_id = request.form.get("user")
  if not user_id:
    flash(_(u"You must provide a user."), 'error')
    return redirect(url_for(".members", community_id=community.slug))
  user_id = int(user_id)
  user = User.query.get(user_id)

  if action in ('add-user-role', 'set-user-role',):
    role = request.form.get("role").lower()

    community.set_membership(user, role)

    if action == 'add-user-role':
      app = current_app._get_current_object()
      activity.send(app, actor=user, verb="join", object=community)

    db.session.commit()
    return redirect(url_for('.members', community_id=community.slug))

  elif action == 'delete':
    membership_id = int(request.form['membership'])
    membership = Membership.query.get(membership_id)
    if membership.user_id != user_id:
      raise InternalServerError()

    community.remove_membership(user)

    app = current_app._get_current_object()
    activity.send(app, actor=user, verb="leave", object=community)

    db.session.commit()
    return redirect(url_for(".members", community_id=community.slug))

  else:
    raise BadRequest('Unknown action: {}'.format(repr(action)))

MEMBERS_EXPORT_HEADERS = [
    _l(u'Name'),
    _l(u'email'),
    _l(u'Last activity in this community'),
    _l(u'Role'),
]

MEMBERS_EXPORT_ATTRS = [
  'User',
  'User.email',
  'last_activity_date',
  'role',
]

HEADER_FONT = openpyxl.styles.Font(bold=True)
HEADER_ALIGN = openpyxl.styles.Alignment(horizontal='center', vertical='top',
                                         wrapText=True)
XLSX_MIME = u'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'

@route("/<string:community_id>/members/excel")
@tab('members')
def members_excel_export():
  community = g.community
  attributes = [attrgetter(a) for a in MEMBERS_EXPORT_ATTRS]
  BaseModel = current_app.db.Model
  wb = openpyxl.Workbook()

  if wb.worksheets:
    wb.remove_sheet(wb.active)

  ws_title = _(u'%(community)s members', community=community.name)
  ws_title = ws_title.strip()
  if len(ws_title) > 31:
    # sheet title cannot exceed 31 char. max length
    ws_title = ws_title[:30] + u'…'
  ws = wb.create_sheet(title=ws_title)
  row = 0
  cells = []

  cols_width = []
  for col, label in enumerate(MEMBERS_EXPORT_HEADERS, 1):
    value = unicode(label)
    cell = WriteOnlyCell(ws, value=value)
    cell.font = HEADER_FONT
    cell.alignment = HEADER_ALIGN
    cells.append(cell)
    cols_width.append(len(value) + 1)

  ws.append(cells)

  for membership_info in _members_query().all():
      row +=1
      cells = []
      for col, getter in enumerate(attributes):
        value = None
        try:
          value = getter(membership_info)
        except AttributeError:
          pass

        if isinstance(value, (BaseModel, Role)):
          value = unicode(value)

        cell = WriteOnlyCell(ws, value=value)
        cells.append(value)

        # estimate width
        value = unicode(cell.value)
        width = max(len(l) for l in value.split(u'\n')) + 1
        cols_width[col] = max(width, cols_width[col])

      ws.append(cells)

  # adjust columns width
  MIN_WIDTH = 3
  MAX_WIDTH = openpyxl.utils.units.BASE_COL_WIDTH * 4

  for idx, width in enumerate(cols_width, 1):
    letter = openpyxl.utils.get_column_letter(idx)
    width = min(max(width, MIN_WIDTH), MAX_WIDTH)
    ws.column_dimensions[letter].width = width

  fd = StringIO.StringIO()
  wb.save(fd)
  fd.seek(0)

  response = current_app.response_class(
    fd,
    mimetype=XLSX_MIME,
  )

  filename = u'{}-members-{}.xlsx'.format(
    community.name,
    strftime("%d:%m:%Y-%H:%M:%S", gmtime())
  )
  response.headers['content-disposition'] = \
      u'attachment;filename="{}"'.format(filename)

  return response

#
# Hack to redirect from urls used by the search engine.
#
@route("/doc/<int:doc_id>")
def doc(doc_id):
  doc = Document.query.get(doc_id)

  if doc is None:
    raise NotFound()

  folder = doc.parent
  while True:
    parent = folder.parent
    if parent.is_root_folder:
      break
    folder = parent
  target_community = Community.query.filter(Community.folder_id == folder.id).one()
  return redirect(url_for("documents.document_view",
                          community_id=target_community.slug, doc_id=doc.id))
